from ...config import Enum, Integer, String, Number, Array, Object


class ModbusType:
    SERIAL = "serial"
    TCP = "tcp"


class ModbusConfig(Object):
    def __init__(self, display_name: str = "Modbus Config"):
        super().__init__(display_name)

        self.type = Enum("Bus Type", choices=["serial", "tcp"], default="serial")
        self.name = String("Name", default="default")

        # todo: only show these if serial type is selected
        self.serial_port = String("Serial Port", default="/dev/ttyAMA0")
        self.serial_baud = Integer("Serial Baud", default=9600)
        # pretty sure this is unused
        self.serial_method = Enum(
            "Serial Method",
            choices=["rtu", "ascii", "socket", "tls"],
            default="rtu",
        )
        self.serial_bits = Integer("Serial Data Bits", default=8)
        self.serial_parity = Enum(
            "Serial Parity", choices=["None", "Even", "Odd"], default="None"
        )
        self.serial_stop = Integer("Serial Stop Bits", default=1)
        self.serial_timeout = Number("Serial Timeout", default=0.3)

        self.tcp_uri = String("TCP URI", default="127.0.0.1:5000")
        self.tcp_timeout = Number("TCP Timeout", default=2.0)


class ManyModbusConfig(Array):
    elements: list[ModbusConfig]

    def __init__(self, display_name: str = "Modbus Config"):
        super().__init__(display_name, element=ModbusConfig("Modbus Instance Config"))



import copy
import inspect
import logging
import re
import time
import json
from datetime import datetime

from typing import Union, Any, Optional, TypeVar, TYPE_CHECKING

from .element import Element
from .interaction import SlimCommand, Interaction, NotSet
from .misc import ApplicationVariant
from .submodule import Container, NAME_VALIDATOR, Application as UIApplication
from .variable import Variable

from ..cloud.api import Client

from ..utils import call_maybe_async, get_is_async, maybe_async, find_object_with_key

if TYPE_CHECKING:
    from ..docker.device_agent.device_agent import DeviceAgentInterface

log = logging.getLogger(__name__)
ElementT = TypeVar("ElementT", bound=Element)
InteractionT = TypeVar("InteractionT", bound=Interaction)


class UIManager:
    def __init__(
        self,
        app_key: str = None,
        client: Union["Client", "DeviceAgentInterface"] = None,
        auto_start: bool = False,
        min_ui_update_period: int = 600,
        min_observed_update_period: int = 4,
        is_async: bool = None,
    ):
        self._is_async = get_is_async(is_async)
        self.client = client
        # to determine whether we can use event-based logic. some reason we can't
        self._has_persistent_connection = client and client.has_persistent_connection()
        self._subscriptions_ready = False

        self.app_key = app_key
        self.app_wrap_ui = True

        self.last_ui_state = (
            dict()
        )  # A python dictionary of the full state from the cloud
        self.last_ui_state_update = None

        self.last_ui_state_wss_connections = dict()
        self.last_ui_state_wss_connections_update = None

        self.last_ui_cmds = dict()
        self.last_ui_cmds_update = None

        self._base_container = UIApplication(name=self.app_key)
        self._interactions: dict[str, Interaction] = dict()

        self._has_critical_interaction_pending: bool = False
        # self._has_critical_ui_state_pending: bool = False
        self._critical_values = dict()  # legacy

        self.min_ui_update_period = min_ui_update_period
        self.min_observed_update_period = min_observed_update_period
        self._last_pushed_time = None
        self._last_push_and_log_time = None

        # legacy, list of subscriptions to call when we have a command update.
        self._cmds_subscriptions = []

        # to allow for @ui.callback("pattern") decorators
        self._command_callbacks = []

        # to keep track of which interactions / ui_cmds to change
        self._changed_interactions = set()

        if auto_start:
            self.start_comms()

    def start_comms(self):
        self._setup_subscriptions()

    def _is_conn_ready(self, setup: bool = False) -> bool:
        if not self._has_persistent_connection:
            return self.client is not None

        if self._subscriptions_ready:
            return True
        elif setup:
            self._setup_subscriptions()
            return self._is_conn_ready(setup=False)  # don't setup for a second time
        else:
            log.error(
                "Attempted use of dda_iface in ui_manager without dda_iface being ready"
            )
            return False

    def is_connected(self) -> bool:
        if not self._has_persistent_connection:
            return self._is_conn_ready()

        if not self._is_conn_ready():
            return False

        return self.client.get_is_dda_online()

    get_is_connected = is_connected

    def is_being_observed(self):
        if not self.last_ui_state_wss_connections:
            return False

        try:
            if not isinstance(self.last_ui_state_wss_connections, dict):
                self.last_ui_state_wss_connections = json.loads(
                    self.last_ui_state_wss_connections
                )
            connections = set(self.last_ui_state_wss_connections["connections"].keys())

            # agent ID will get set on startup in processor Client and on config load in apps.
            if self.client.agent_id and len(connections - {self.client.agent_id}) > 0:
                return True

            # if there is more than one connection, then we are being observed
            return len(connections) > 1
        except Exception as e:
            log.error(f"Error checking if being observed: {e}")
            return False

    def has_been_connected(self):
        if self._has_persistent_connection and self.client is not None:
            return self.client.get_has_dda_been_online()
        return self.last_ui_state is not None

    get_has_been_connected = has_been_connected

    def _setup_subscriptions(self):
        if not self._has_persistent_connection:
            log.error(
                "Attempted to setup subscriptions without valid connection client."
            )
            return

        if self.client is None:
            log.warning("Attempted to setup subscriptions without client being set")
            return

        log.info("Setting up dda subscriptions")
        self.client.add_subscription("ui_state", self.on_state_update)
        self.client.add_subscription(
            "ui_state@wss_connections", self.on_state_wss_update
        )
        self.client.add_subscription("ui_cmds", self.on_command_update_async)

        self._subscriptions_ready = True

    async def await_comms_sync(self, timeout: int = 5) -> bool:
        if not self._has_persistent_connection:
            return False

        if not hasattr(self.client, "wait_for_channels_sync_async"):
            log.error("Attempted to await comms sync without valid connection client.")
            return False
        return await self.client.wait_for_channels_sync_async(
            channel_names=["ui_state", "ui_cmds"],
            timeout=timeout,
        )

    async def on_state_update(self, _, aggregate: dict[str, Any]):
        self._set_new_ui_state(aggregate)

    async def on_state_wss_update(self, _, aggregate: dict[str, Any]):
        self.last_ui_state_wss_connections = aggregate
        self.last_ui_state_wss_connections_update = time.time()

    def _on_command_update_common(self, aggregate: dict[str, Any]):
        aggregate = self._set_new_ui_cmds(aggregate)

        # add commands that don't currently exist
        to_add = {k: v for k, v in aggregate.items() if k not in self._interactions}
        for name, current_value in to_add.items():
            self._interactions[name] = SlimCommand(name, current_value)

        return aggregate

    def on_command_update(self, _, aggregate: dict[str, Any]):
        log.debug(f"running on command update, {aggregate}")
        self._on_command_update_common(aggregate)

        changed = {c: v for c, v in aggregate.items()}

        ## Iterate through all the commands that we have locally, and call the callback if it exists
        for command_name, command in self._interactions.items():
            if command_name in changed:
                new_value = changed[command_name]
            else:
                new_value = None
                if hasattr(command, "default") and command.default is not None:
                    new_value = command.default

            if not command._is_new_value(new_value):
                continue

            command._handle_new_value(new_value)

            for pattern, callback in self._command_callbacks:
                if pattern.match(command_name):
                    callback(command, new_value)

    async def on_command_update_async(self, _, aggregate: dict[str, Any]):
        log.debug("Running on_command_update_async")
        prev_agg = copy.deepcopy(self.last_ui_cmds)
        aggregate = self._on_command_update_common(aggregate)

        # call all subscribed to cmds updates
        for c in self._cmds_subscriptions:
            log.debug(f"Invoking command subscription: {c}")
            await call_maybe_async(c)

        # work out command diff and call individual commands
        changed = {c: v for c, v in aggregate.items()}

        log.debug(f"Prev Agg ({self.last_ui_cmds_update}): {prev_agg}")
        log.debug(f"New/Current Agg: {aggregate}")
        # log.debug(f"Changed: {changed}")

        ## Iterate through all the commands that we have locally, and call the callback if it exists
        for command_name, command in self._interactions.items():
            if command_name in changed:
                new_value = changed[command_name]
            else:
                new_value = None
                if hasattr(command, "default") and command.default is not None:
                    new_value = command.default

            if not command._is_new_value(new_value):
                continue

            await command._handle_new_value_async(new_value)

            for pattern, callback in self._command_callbacks:
                if pattern.match(command_name):
                    await call_maybe_async(callback, command, new_value)

    def _set_new_ui_cmds(self, payload: dict[str, Any]):
        if not payload:
            log.info("Received empty UI commands payload.")
            payload = {}

        if isinstance(payload, str):
            try:
                payload = json.loads(payload)
            except json.JSONDecodeError:
                log.error(f"Failed to decode UI commands: {payload}")
                payload = {}

        try:
            payload = payload["cmds"]
        except KeyError:
            pass

        self.last_ui_cmds = copy.deepcopy(payload)
        self.last_ui_cmds_update = time.time()
        return payload

    def _set_new_ui_state(self, payload: dict[str, Any]):
        if not isinstance(payload, dict):
            payload = {}

        try:
            payload = payload["state"]
        except KeyError:
            pass

        try:
            payload = payload["children"][self.app_key]
        except KeyError:
            payload = {}

        self.last_ui_state = payload
        self.last_ui_state_update = time.time()

        ## TODO: Implement this ????
        # if self._base_container is not None:
        #     self._base_container.from_dict(payload)

        ## Iterate through the payload and update anything that needs updating
        # define a function to recursively trawl through each element of the last ui state and allow the element to update itself
        def update_elements_from_ui_state(d):
            for k, v in d.items():
                if isinstance(v, dict):
                    element = self.get_element(k)
                    if element is not None:
                        element.recv_ui_state_update(v)
                    else:
                        update_elements_from_ui_state(v)

        update_elements_from_ui_state(payload)

    def get_full_interaction_key(self, name: str) -> str:
        # Get the complete name of the key for the interaction
        if self.app_key in name:
            return name
        return f"{self.app_key}_{name.strip()}"

    def _transform_interaction_name(self, name):
        # inject the app key (unique) into the interaction name
        # so we don't have namespace collisions between apps.
        if self.app_key in name:
            return name
        return f"{self.app_key}_{name.strip()}"

    def _add_interaction(self, interaction: Interaction):
        # inject the app key (unique) into the interaction name
        # so we don't have namespace collisions between apps.
        name = self._transform_interaction_name(interaction.name)
        interaction.name = name

        if not NAME_VALIDATOR.match(name):
            raise RuntimeError(
                f"Invalid name '{name}' for interaction '{interaction}'. "
                f"Valid characters include letters, numbers, and underscores."
            )

        if name in self._interactions:
            ## If the interaction already exists, we should preserve the current value if it exists
            if (
                hasattr(self._interactions[name], "_current_value")
                and self._interactions[name]._current_value is not NotSet
            ):
                interaction.current_value = self._interactions[name]._current_value

        self._interactions[name] = interaction
        interaction._manager = self

    def _remove_interaction(self, interaction_name: str) -> None:
        try:
            del self._interactions[interaction_name]
        except KeyError:
            return

    def add_interaction(self, interaction: InteractionT):
        if not isinstance(interaction, Interaction) and hasattr(
            interaction, "_ui_type"
        ):
            interaction = self._register_interaction(interaction, interaction.__self__)

        self._add_interaction(interaction)
        self._base_container.add_children(interaction)

    @staticmethod
    def _register_interaction(func, parent) -> InteractionT:
        item = func._ui_type(**func._ui_kwargs)
        item.callback = func

        try:
            setattr(parent, func.__name__, item)
        except AttributeError:
            pass  # maybe they've initialised the function w/o a class

        return item

    def register_interactions(self, obj_to_search):
        for name, func in inspect.getmembers(
            obj_to_search,
            predicate=lambda f: inspect.ismethod(f) and hasattr(f, "_ui_type"),
        ):
            self._register_interaction(func, obj_to_search)

    def register_callbacks(self, obj_to_search):
        for name, func in inspect.getmembers(
            obj_to_search,
            predicate=lambda f: inspect.ismethod(f) and hasattr(f, "_is_ui_callback"),
        ):
            if isinstance(func._ui_callback_pattern, str):
                p = func._ui_callback_pattern
                if func._is_ui_global_interaction:
                    pattern = re.compile(p)
                else:
                    pattern = re.compile(self._transform_interaction_name(p))

            elif isinstance(func._ui_callback_pattern, re.Pattern):
                pattern = func._ui_callback_pattern
                if func._is_ui_global_interaction is False:
                    pattern = re.compile(
                        self._transform_interaction_name(pattern.pattern)
                    )

            else:
                raise ValueError(
                    "Invalid pattern type for UI callback. Must be a string or re.Pattern."
                )

            log.info(f"Registering UI callback ({func}) with pattern {pattern.pattern}")
            self._command_callbacks.append((pattern, func))

    def get_interaction(self, name: str) -> Optional[Interaction]:
        try:
            return self._interactions[self._transform_interaction_name(name)]
        except KeyError:
            return None

    def update_interaction(self, name: str, updated: Interaction) -> bool:
        name = self._transform_interaction_name(name)
        if name not in self._interactions:
            return False

        self._interactions[name] = updated
        return True

    add_command = add_interaction
    get_command = get_interaction

    def coerce_command(
        self, command_name: str, value: Any, critical: bool = False
    ) -> None:
        command = self.get_command(command_name)
        if command is None:
            log.info(f"Tried to coerce command {command_name} that doesn't exist.")
            return

        command.coerce(value, critical=critical)

    def get_element(self, element_name: str) -> Optional[ElementT]:
        return self._base_container.get_element(element_name)

    def get_from_ui_state(self, element_name: str) -> Optional[dict]:
        return find_object_with_key(self.last_ui_state, element_name)

    def update_variable(
        self, variable_name: str, value: Any, critical: bool = False
    ) -> bool:
        element = self._base_container.get_element(variable_name)
        if not (element and isinstance(element, Variable)):
            log.info(
                f"PyDoover: Tried to update variable '{variable_name}' that doesn't exist."
            )
            return False

        if critical is True and element.current_value != value:
            self._has_critical_interaction_pending = True

        element.current_value = value
        return True

    def add_cmds_update_subscription(self, callback):
        # fixme: create alias or something
        self._cmds_subscriptions.append(callback)

    def get_all_interactions(self) -> list[Interaction]:
        return list(self._interactions.values())

    def get_all_interaction_names(self) -> list[str]:
        return list(self._interactions.keys())

    def get_all_variables(self) -> list[Variable]:
        if self._base_container is None:
            return []
        return self._base_container.get_all_elements(type_filter=Variable)

    get_available_commands = get_all_interaction_names

    def record_critical_value(self, name, value):
        log.warning(
            "this function is deprecated. use the critical=True parameter of another appropriate function."
        )
        if self._critical_values.get(name) == value:
            return

        self._critical_values[name] = value
        self._has_critical_interaction_pending = True
        # self._has_critical_ui_state_pending = True

    def _assess_pending_log_requests(self):
        ## Iterate through all variables and check if they have a pending log request
        log_requested = False
        for variable in self.get_all_variables():
            if variable.has_pending_log_request():
                log.debug(f"Variable {variable.name} has a pending log request")
                variable.clear_log_request()
                log_requested = True
        return log_requested

    def _get_max_age(self, force_update: bool = False) -> tuple[bool, int]:
        period = (
            self.min_observed_update_period
            if self.is_being_observed()
            else self.min_ui_update_period
        )

        if force_update:
            return True, 1

        if self._has_critical_interaction_pending:
            return True, 1
        if self._last_pushed_time is None or self._last_push_and_log_time is None:
            return True, 1

        since_last_push_and_log = time.time() - self._last_push_and_log_time
        if since_last_push_and_log > self.min_ui_update_period:
            return True, period

        return False, period

    @maybe_async()
    def handle_comms(self, force_log: bool = False):
        log_requested = self._assess_pending_log_requests()
        force_log, max_age = self._get_max_age(force_log or log_requested)
        self.push(force_log, max_age)

    async def handle_comms_async(self, force_log: bool = False):
        log_requested = self._assess_pending_log_requests()
        force_log, max_age = self._get_max_age(force_log or log_requested)
        await self.push_async(force_log, max_age)

    @maybe_async()
    def send_notification(self, message: str, record_activity: bool = True):
        self.publish_to_channel(
            "significantEvent",
            {
                "notification_msg": message,
            },
            record_log=True,
            max_age=1,
        )
        if record_activity:
            self.record_activity(message)

    async def send_notification_async(self, message: str, record_activity: bool = True):
        await self._publish_to_channel_async(
            "significantEvent",
            {
                "notification_msg": message,
            },
            record_log=True,
            max_age=1,
        )
        if record_activity:
            await self.record_activity_async(message)

    async def record_activity_async(self, message: str):
        await self._publish_to_channel_async(
            "activity_log",
            {
                "action_string": message,
            },
            record_log=True,
        )

    def publish_to_channel(
        self,
        channel_name: str,
        data: dict[str, Any],
        record_log: bool = True,
        max_age: int = None,
        timestamp: Optional[datetime] = None,
        **kwargs,
    ):
        return self._publish_to_channel(
            channel_name,
            data,
            record_log=record_log,
            timestamp=timestamp,
            max_age=max_age,
            **kwargs,
        )

    def _publish_to_channel(
        self,
        channel_name: str,
        data: dict[str, Any],
        record_log: bool = True,
        timestamp: Optional[datetime] = None,
        max_age: int = None,
        **kwargs,
    ):
        # this purely exists to provide cross-compatibility between clients (hence private method).
        if isinstance(self.client, Client):
            channel = self.client.get_channel_named(channel_name, self.agent_id)
            return channel.publish(
                data, save_log=record_log, timestamp=timestamp, **kwargs
            )
        else:
            # fixme: allow for timestamp in DDA message publishing...
            return self.client.publish_to_channel(
                channel_name, data, record_log=record_log, max_age=max_age
            )

    async def _publish_to_channel_async(
        self,
        channel_name: str,
        data: dict[str, Any],
        record_log: bool = True,
        timestamp: Optional[datetime] = None,
        max_age: int = None,
        **kwargs,
    ):
        if isinstance(self.client, Client):
            # in theory this works but lets just discourage this behaviour...
            raise RuntimeError("Cannot push async with a Client object")

        return await self.client.publish_to_channel_async(
            channel_name, data, record_log=record_log, max_age=max_age
        )

    @maybe_async()
    def pull(self):
        print("pulling...")
        if isinstance(self.client, Client):
            ui_cmds = self.client.get_channel_named("ui_cmds", self.agent_id)
            ui_state = self.client.get_channel_named("ui_state", self.agent_id)

            ui_cmds_agg = ui_cmds.fetch_aggregate()
            ui_state_agg = ui_state.fetch_aggregate()
        else:
            ui_cmds_agg = self.client.get_channel_aggregate("ui_cmds")
            ui_state_agg = self.client.get_channel_aggregate("ui_state")

        self._set_new_ui_state(ui_state_agg)

        # self._set_new_ui_cmds(ui_cmds_agg)
        try:
            ui_cmds_agg = ui_cmds_agg["cmds"]
        except Exception as e:
            log.warning(f"Failed to get UI commands: {e}")
            ui_cmds_agg = {}

        self.on_command_update(None, ui_cmds_agg)

    async def pull_async(self):
        if isinstance(self.client, Client):
            raise RuntimeError("Cannot pull async with a Client object")

        ui_cmds_agg = await self.client.get_channel_aggregate_async("ui_cmds")
        ui_state_agg = await self.client.get_channel_aggregate_async("ui_state")
        self._set_new_ui_state(ui_state_agg)
        # self._set_new_ui_cmds(ui_cmds_agg)
        await self.on_command_update_async(None, ui_cmds_agg)

    def _check_dda_ready(self):
        if not self._is_conn_ready():
            log.warning("Attempted to push config without ready connection client.")
            return False
        elif not self.client.get_has_dda_been_online():
            # for a persistent connection, don't push if we haven't first pulled last data
            # HTTP-based connections will do a pull before pushing so that is fine.
            log.warning("Attempted to push config without DDA being online.")
            return False
        elif self.last_ui_state_update is None:
            log.warning("Waiting for UI state update to be pulled before pushing...")
            return False
        elif self.last_ui_cmds_update is None:
            log.warning("Waiting for UI commands to be pulled before pushing...")
            return False
        return True

    def _wrap_ui_state(
        self, data: dict[str, Any] | None, clear: bool = False
    ) -> dict[str, Any]:
        if clear is False and not data:
            data = {}

        if self.app_wrap_ui:
            return {"state": {"children": {self.app_key: data}}}
        else:
            return {"state": data}

    @maybe_async()
    def push(
        self,
        record_log: bool = True,
        max_age: Optional[int] = None,
        should_remove: bool = True,
        timestamp: Optional[datetime] = None,
        even_if_empty: bool = False,
        only_channels: Optional[list] = None,
        publish_fields: Optional[list] = None,
    ) -> bool:
        publish_fields = publish_fields or []

        # self.check_dda()
        if self._has_persistent_connection:
            if not self._check_dda_ready():
                return False
        else:
            self.pull()  # do a pull before HTTP client pushes anything...

        print("pushing sync...")

        commands_update = self._get_commands_update(publish_fields=publish_fields)
        if commands_update is not None and (
            only_channels is None or "ui_cmds" in only_channels
        ):
            self._publish_to_channel(
                "ui_cmds",
                {"cmds": commands_update},
                timestamp=timestamp,
                max_age=1,
                record_log=True,
            )

        ui_state_update = self._get_ui_state_update(
            should_remove=should_remove, retain_fields=publish_fields
        )
        if even_if_empty or (
            ui_state_update is not None
            and (only_channels is None or "ui_state" in only_channels)
        ):
            self._publish_to_channel(
                "ui_state",
                self._wrap_ui_state(ui_state_update),
                record_log=record_log,
                timestamp=timestamp,
                max_age=max_age,
            )
        else:
            print("not pushing empty ui state")

        self._last_pushed_time = time.time()
        if record_log:
            self._last_push_and_log_time = time.time()

        self._has_critical_interaction_pending = False
        return True

    async def push_async(
        self,
        record_log: bool = True,
        max_age: Optional[int] = None,
        should_remove: bool = True,
        timestamp: Optional[datetime] = None,
        even_if_empty: bool = False,
        only_channels: Optional[list] = None,
        publish_fields: Optional[list] = None,
    ) -> bool:
        publish_fields = publish_fields or []

        # self.check_dda()
        if self._has_persistent_connection:
            if not self._check_dda_ready():
                return False
        else:
            self.pull()  # do a pull before HTTP client pushes anything...

        commands_update = self._get_commands_update(publish_fields=publish_fields)
        if commands_update is not None and (
            only_channels is None or "ui_cmds" in only_channels
        ):
            log.debug("Pushing UI commands")
            await self._publish_to_channel_async(
                "ui_cmds",
                {"cmds": commands_update},
                timestamp=timestamp,
                max_age=1,
                record_log=True,
            )

        ui_state_update = self._get_ui_state_update(
            should_remove=should_remove, retain_fields=publish_fields
        )
        if even_if_empty or (
            ui_state_update is not None
            and (only_channels is None or "ui_state" in only_channels)
        ):
            log.debug("Pushing UI state")
            await self._publish_to_channel_async(
                "ui_state",
                self._wrap_ui_state(ui_state_update),
                record_log=record_log,
                timestamp=timestamp,
                max_age=max_age,
            )

        self._last_pushed_time = time.time()
        if record_log:
            self._last_push_and_log_time = time.time()

        self._has_critical_interaction_pending = False
        return True

    @maybe_async()
    def clear_ui(self):
        # this could be dangerous...
        log.info("Clearing UI")
        self._publish_to_channel(
            "ui_state",
            self._wrap_ui_state(None, clear=True),
            max_age=-1,
            record_log=False,
        )

    async def clear_ui_async(self):
        log.info("Clearing UI")
        # max-age=-1 to force an update right now. otherwise, first update will remove effect of "clearing" data.
        await self._publish_to_channel_async(
            "ui_state", self._wrap_ui_state(None, clear=True), max_age=-1
        )

    def _get_commands_update(
        self,
        publish_fields: list | None = None,
        only_local_changes: bool = True,
    ) -> dict[str, Any] | None:
        if publish_fields is None:
            publish_fields = []

        cloud_commands = copy.deepcopy(self.last_ui_cmds)
        local_commands = {
            k: v._json_safe_current_value() for k, v in self._interactions.items()
        }

        # don't include commands that are the same as the cloud, and values that aren't set
        # don't override other apps' commands if we haven't changed the value.
        result = {
            name: value
            for name, value in local_commands.items()
            if (
                cloud_commands.get(name) != value
                and value != NotSet
                and name in self._changed_interactions
                if only_local_changes
                else True
            )
            or name in publish_fields
        }

        self._changed_interactions.clear()

        # don't clean up commands that exist upstream but not locally for now.
        result.update(
            {c: None for c in cloud_commands.keys() if c not in local_commands}
        )

        log.debug("Last Commands: " + str(cloud_commands))
        log.debug("New Commands: " + str(local_commands))
        log.debug("Commands Update: " + str(result))

        if len(result) == 0:
            return None
        return result

    def _get_ui_state_update(
        self, should_remove: bool = True, retain_fields: list = list
    ) -> Optional[dict[str, Any]]:
        cloud_state = self.last_ui_state or {}
        # this recursively evaluates and finds the diff on all children, rather than trying to do the diff here
        result = self._base_container.get_diff(
            cloud_state, remove=should_remove, retain_fields=retain_fields
        )

        log.debug("Last UI State: " + str(cloud_state))
        log.debug("New UI State: " + str(self._base_container.to_dict()))
        log.debug("UI State Update: " + str(result))

        if not result or len(result) == 0:
            return None

        return result

    def _maybe_add_interaction_from_elems(
        self, *elements: Union[Element, Container]
    ) -> list[Element]:
        to_return = []
        for element in elements:
            # this is a bit hacky, but it's to stop passing in an unregistered interaction (ie. created with a decorator
            # outside of a submodule and hasn't been registered yet),
            # instead we'll silently register it and proceed as-is
            if not isinstance(element, Interaction) and hasattr(element, "_ui_type"):
                element = self._register_interaction(element, element.__self__)

            if isinstance(element, Container):
                self._maybe_add_interaction_from_elems(*element.children)
            elif isinstance(element, Interaction):
                self._add_interaction(element)
            to_return.append(element)

        return to_return

    def add_children(self, *children: Element) -> None:
        if len(children) == 1 and isinstance(children[0], list):
            # for backwards compatibility, this used to accept a single list of children
            children = children[0]

        updated = self._maybe_add_interaction_from_elems(*children)
        self._base_container.add_children(*updated)

    def remove_children(self, *children: Element) -> None:
        if len(children) == 1 and isinstance(children[0], list):
            # for backwards compatibility, this used to accept a single list of children
            children = children[0]

        for elem in children:
            if not isinstance(elem, Element):
                # sometimes an unregistered function can end up here and break things...
                continue

            if elem == self._base_container:
                raise RuntimeError("You can't remove the base container!")

            # this should never be None, but in case some numpty does something weird...
            if getattr(elem, "parent", None):
                elem.parent.remove_children(elem)

            ## Remove the element
            self._base_container.remove_children(elem)
            self._remove_interaction(elem.name)

    def set_children(self, children: list[Element]) -> None:
        updated = self._maybe_add_interaction_from_elems(*children)
        self._base_container.set_children(updated)
        # self._maybe_add_interaction_from_elems(*children)
        # self._base_container.set_children(children)

        # self._base_container.add_children( self.cameras )
        # if len(self.cameras) > 0:
        #     self._base_container.add_children( [ doover_ui_hidden_value(name="last_cam_snapshot") ] )

    def set_status_icon(self, icon_type: str, critical: bool = False):
        if icon_type == self._base_container.status_icon:
            return
        elif critical is True:
            # self._has_critical_ui_state_pending = True
            self._has_critical_interaction_pending = (
                True  # fixme: work out if we ever need this in element setting
            )

        self._base_container.status_icon = icon_type

    def set_display_name(self, name: str, critical: bool = False):
        if name == self._base_container.display_name:
            return
        elif critical is True:
            self._has_critical_interaction_pending = True

        self._base_container.display_name = name

    def set_variant(self, variant: ApplicationVariant) -> None:
        if variant not in (ApplicationVariant.submodule, ApplicationVariant.stacked):
            raise ValueError("Variant must be one of 'stacked', 'submodule'")

        self._base_container.variant = str(variant)

    def set_position(self, position: int):
        self._base_container.position = position



from typing import Union
from datetime import datetime

from .interaction import Interaction
from .misc import NotSet


class Parameter(Interaction):
    """Base class for UI parameters. All parameters should inherit from this class."""

    type = NotImplemented


class NumericParameter(Parameter):
    """A numeric parameter that can be used to represent both integer and float values.

    Attributes
    ----------
    name: str
        The name of the parameter.
    display_name: str
        The display name of the parameter.
    min: Union[int, float], optional
        The minimum value for the parameter. Defaults to None.
    max: Union[int, float], optional
        The maximum value for the parameter. Defaults to None.
    """

    type = "uiFloatParam"

    def __init__(
        self,
        name,
        display_name,
        min_val: Union[int, float] = None,
        max_val: Union[int, float] = None,
        **kwargs,
    ):
        super().__init__(name, display_name, **kwargs)

        self.min = min_val or kwargs.pop("float_min", None)
        self.max = max_val or kwargs.pop("float_max", None)

    def to_dict(self):
        result = super().to_dict()

        if self.min is not None:
            result["min"] = self.min
        if self.max is not None:
            result["max"] = self.max

        return result


class TextParameter(Parameter):
    """A text parameter that can be used to represent string values for a user to modify.

    This can either be inline or a large form text area, depending on the `is_text_area` parameter.

    Attributes
    ----------
    name: str
        The name of the parameter.
    display_name: str
        The display name of the parameter.
    is_text_area: bool
        Whether the text parameter should be displayed as a text area. Defaults to False.
    """

    type = "uiTextParam"

    def __init__(self, name, display_name, is_text_area: bool = False, **kwargs):
        super().__init__(name, display_name, **kwargs)
        self.is_text_area = is_text_area

    def to_dict(self):
        result = super().to_dict()
        result["isTextArea"] = self.is_text_area
        return result


class BooleanParameter(Parameter):
    """
    A boolean parameter that can be used to represent true/false values.

    Warnings
    --------
    This is not implemented in the doover site yet, and will raise a NotImplementedError if used!

    Attributes
    ----------
    name: str
        The name of the parameter.
    display_name: str
        The display name of the parameter.
    """

    type = "uiBoolParam"

    def __init__(self, name, display_name, **kwargs):
        super().__init__(name, display_name, **kwargs)
        raise NotImplementedError("boolean parameter not implemented in doover site.")


class DateTimeParameter(Parameter):
    """Date and time parameter that can be used to request date/time values from a user.

    Internally, all datetime values are stored as epoch seconds in UTC

    Attributes
    ----------
    name: str
        The name of the parameter.
    display_name: str
        The display name of the parameter.
    include_time: bool
        Whether to include time in the datetime picker. Defaults to False.
    """

    type = "uiDatetimeParam"

    def __init__(
        self, name: str, display_name: str, include_time: bool = False, **kwargs
    ):
        super().__init__(name, display_name, **kwargs)
        self.include_time = include_time

    @property
    def current_value(self) -> datetime | None:
        """datetime, optional: Returns the current value of the parameter as a datetime object, or `None` if it isn't set."""
        if self._current_value is NotSet or self._current_value is None:
            return None
        if isinstance(self._current_value, datetime):
            return self._current_value
        elif isinstance(self._current_value, (int, float)):
            return datetime.utcfromtimestamp(self._current_value)
        return None

    @current_value.setter
    def current_value(self, new_val):
        if isinstance(new_val, datetime):
            new_val = int(new_val.timestamp())
        self._current_value = new_val

    def to_dict(self):
        result = super().to_dict()
        result["includeTime"] = self.include_time
        return result


def numeric_parameter(
    name: str,
    display_name: str,
    min_val: Union[int, float] = None,
    max_val: Union[int, float] = None,
    **kwargs,
):
    """Decorator to create a numeric parameter for a function.

    The function decorated by this decorator will be called whenever the value of the numeric parameter changes.

    Examples
    --------

    A basic numeric parameter with a range of 0 to 100 ::

        @ui.numeric_parameter(
            name="example_numeric",
            display_name="Example Numeric Parameter",
            min_val=0,
            max_val=100
        )
        def example_function(value: float):
            print(f"Numeric parameter changed to: {value}")


    Parameters
    ----------
    name: str
        The name of the parameter.
    display_name: str
        The display name of the parameter.
    min_val: Union[int, float], optional
        The minimum value for the parameter. Defaults to None.
    max_val: Union[int, float], optional
        The maximum value for the parameter. Defaults to None.
    """

    def decorator(func):
        func._ui_type = NumericParameter
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "min_val": min_val,
            "max_val": max_val,
            **kwargs,
        }
        return func

    return decorator


def text_parameter(name: str, display_name: str, is_text_area: bool = False, **kwargs):
    """Decorator to create a text parameter for a function.

    The function decorated by this decorator will be called whenever the value of the text parameter changes.

    Examples
    --------

    A basic text parameter ::

        @ui.text_parameter(
            name="example_text",
            display_name="Example Text Parameter"
        )
        def example_function(value: str):
            print(f"Text parameter changed to: {value}")


    Parameters
    ----------
    name: str
        The name of the parameter.
    display_name: str
        The display name of the parameter.
    is_text_area: bool
        Whether the text parameter should be displayed as a text area. Defaults to False.
    """

    def decorator(func):
        func._ui_type = TextParameter
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "is_text_area": is_text_area,
            **kwargs,
        }
        return func

    return decorator


def boolean_parameter(name: str, display_name: str, **kwargs):
    """Decorator to create a boolean parameter for a function.

    The function decorated by this decorator will be called whenever the value of the boolean parameter changes.

    Warnings
    --------
    This is not implemented in the doover site yet, and will raise a NotImplementedError if used!

    Parameters
    ----------
    name: str
        The name of the parameter.
    display_name: str
        The display name of the parameter.
    """

    def decorator(func):
        func._ui_type = BooleanParameter
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            **kwargs,
        }
        return func

    return decorator


def datetime_parameter(
    name: str, display_name: str, include_time: bool = False, **kwargs
):
    """Decorator to create a datetime parameter for a function.

    The function decorated by this decorator will be called whenever the value of the datetime parameter (picker) changes.

    Examples
    --------
    A basic datetime parameter with time included ::

        @ui.datetime_parameter(
            name="example_datetime",
            display_name="Example Datetime Parameter",
            include_time=True
        )
        def example_function(value: datetime):
            print(f"Datetime parameter changed to: {value}")


    Parameters
    ----------
    name: str
        The name of the parameter.
    display_name: str
        The display name of the parameter.
    include_time: bool
        Whether to include time in the datetime picker. Defaults to False.
    """

    def decorator(func):
        func._ui_type = DateTimeParameter
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "include_time": include_time,
            **kwargs,
        }
        return func

    return decorator



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class busStatus(_message.Message):
    __slots__ = ("bus_id", "open", "serial_settings", "ethernet_settings")
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    OPEN_FIELD_NUMBER: _ClassVar[int]
    SERIAL_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    ETHERNET_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    bus_id: str
    open: bool
    serial_settings: serialBusSettings
    ethernet_settings: ethernetBusSettings
    def __init__(self, bus_id: _Optional[str] = ..., open: bool = ..., serial_settings: _Optional[_Union[serialBusSettings, _Mapping]] = ..., ethernet_settings: _Optional[_Union[ethernetBusSettings, _Mapping]] = ...) -> None: ...

class responseHeader(_message.Message):
    __slots__ = ("success", "response_code", "response_message", "bus_status")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_CODE_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    BUS_STATUS_FIELD_NUMBER: _ClassVar[int]
    success: bool
    response_code: int
    response_message: str
    bus_status: _containers.RepeatedCompositeFieldContainer[busStatus]
    def __init__(self, success: bool = ..., response_code: _Optional[int] = ..., response_message: _Optional[str] = ..., bus_status: _Optional[_Iterable[_Union[busStatus, _Mapping]]] = ...) -> None: ...

class serverResponseHeader(_message.Message):
    __slots__ = ("success", "response_code", "response_message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_CODE_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    response_code: int
    response_message: str
    def __init__(self, success: bool = ..., response_code: _Optional[int] = ..., response_message: _Optional[str] = ...) -> None: ...

class testCommsRequest(_message.Message):
    __slots__ = ("message",)
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class testCommsResponse(_message.Message):
    __slots__ = ("response_header", "response")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    response_header: responseHeader
    response: str
    def __init__(self, response_header: _Optional[_Union[responseHeader, _Mapping]] = ..., response: _Optional[str] = ...) -> None: ...

class serialBusSettings(_message.Message):
    __slots__ = ("port", "baud", "modbus_method", "data_bits", "parity", "stop_bits", "timeout")
    PORT_FIELD_NUMBER: _ClassVar[int]
    BAUD_FIELD_NUMBER: _ClassVar[int]
    MODBUS_METHOD_FIELD_NUMBER: _ClassVar[int]
    DATA_BITS_FIELD_NUMBER: _ClassVar[int]
    PARITY_FIELD_NUMBER: _ClassVar[int]
    STOP_BITS_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_FIELD_NUMBER: _ClassVar[int]
    port: str
    baud: int
    modbus_method: str
    data_bits: int
    parity: str
    stop_bits: int
    timeout: float
    def __init__(self, port: _Optional[str] = ..., baud: _Optional[int] = ..., modbus_method: _Optional[str] = ..., data_bits: _Optional[int] = ..., parity: _Optional[str] = ..., stop_bits: _Optional[int] = ..., timeout: _Optional[float] = ...) -> None: ...

class ethernetBusSettings(_message.Message):
    __slots__ = ("ip", "port", "modbus_method", "timeout")
    IP_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    MODBUS_METHOD_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_FIELD_NUMBER: _ClassVar[int]
    ip: str
    port: int
    modbus_method: str
    timeout: float
    def __init__(self, ip: _Optional[str] = ..., port: _Optional[int] = ..., modbus_method: _Optional[str] = ..., timeout: _Optional[float] = ...) -> None: ...

class openBusRequest(_message.Message):
    __slots__ = ("bus_id", "serial_settings", "ethernet_settings")
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    SERIAL_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    ETHERNET_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    bus_id: str
    serial_settings: serialBusSettings
    ethernet_settings: ethernetBusSettings
    def __init__(self, bus_id: _Optional[str] = ..., serial_settings: _Optional[_Union[serialBusSettings, _Mapping]] = ..., ethernet_settings: _Optional[_Union[ethernetBusSettings, _Mapping]] = ...) -> None: ...

class openBusResponse(_message.Message):
    __slots__ = ("response_header", "bus_id")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    response_header: responseHeader
    bus_id: str
    def __init__(self, response_header: _Optional[_Union[responseHeader, _Mapping]] = ..., bus_id: _Optional[str] = ...) -> None: ...

class listBusRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class listBusResponse(_message.Message):
    __slots__ = ("response_header", "bus_status")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    BUS_STATUS_FIELD_NUMBER: _ClassVar[int]
    response_header: responseHeader
    bus_status: _containers.RepeatedCompositeFieldContainer[busStatus]
    def __init__(self, response_header: _Optional[_Union[responseHeader, _Mapping]] = ..., bus_status: _Optional[_Iterable[_Union[busStatus, _Mapping]]] = ...) -> None: ...

class busStatusRequest(_message.Message):
    __slots__ = ("bus_id",)
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    bus_id: str
    def __init__(self, bus_id: _Optional[str] = ...) -> None: ...

class busStatusResponse(_message.Message):
    __slots__ = ("response_header", "bus_status")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    BUS_STATUS_FIELD_NUMBER: _ClassVar[int]
    response_header: responseHeader
    bus_status: busStatus
    def __init__(self, response_header: _Optional[_Union[responseHeader, _Mapping]] = ..., bus_status: _Optional[_Union[busStatus, _Mapping]] = ...) -> None: ...

class closeBusRequest(_message.Message):
    __slots__ = ("bus_id",)
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    bus_id: str
    def __init__(self, bus_id: _Optional[str] = ...) -> None: ...

class closeBusResponse(_message.Message):
    __slots__ = ("response_header", "bus_id")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    response_header: responseHeader
    bus_id: str
    def __init__(self, response_header: _Optional[_Union[responseHeader, _Mapping]] = ..., bus_id: _Optional[str] = ...) -> None: ...

class readRegisterRequest(_message.Message):
    __slots__ = ("bus_id", "modbus_id", "register_type", "address", "count")
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    MODBUS_ID_FIELD_NUMBER: _ClassVar[int]
    REGISTER_TYPE_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    bus_id: str
    modbus_id: int
    register_type: int
    address: int
    count: int
    def __init__(self, bus_id: _Optional[str] = ..., modbus_id: _Optional[int] = ..., register_type: _Optional[int] = ..., address: _Optional[int] = ..., count: _Optional[int] = ...) -> None: ...

class readRegisterResponse(_message.Message):
    __slots__ = ("response_header", "response_code", "values")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_CODE_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    response_header: responseHeader
    response_code: int
    values: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, response_header: _Optional[_Union[responseHeader, _Mapping]] = ..., response_code: _Optional[int] = ..., values: _Optional[_Iterable[int]] = ...) -> None: ...

class writeRegisterRequest(_message.Message):
    __slots__ = ("bus_id", "modbus_id", "register_type", "address", "values")
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    MODBUS_ID_FIELD_NUMBER: _ClassVar[int]
    REGISTER_TYPE_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    bus_id: str
    modbus_id: int
    register_type: int
    address: int
    values: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, bus_id: _Optional[str] = ..., modbus_id: _Optional[int] = ..., register_type: _Optional[int] = ..., address: _Optional[int] = ..., values: _Optional[_Iterable[int]] = ...) -> None: ...

class writeRegisterResponse(_message.Message):
    __slots__ = ("response_header", "response_code")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_CODE_FIELD_NUMBER: _ClassVar[int]
    response_header: responseHeader
    response_code: int
    def __init__(self, response_header: _Optional[_Union[responseHeader, _Mapping]] = ..., response_code: _Optional[int] = ...) -> None: ...

class scheduleWriteRegisterRequest(_message.Message):
    __slots__ = ("bus_id", "register_type", "address", "values", "delay_secs")
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    REGISTER_TYPE_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    DELAY_SECS_FIELD_NUMBER: _ClassVar[int]
    bus_id: str
    register_type: int
    address: int
    values: _containers.RepeatedScalarFieldContainer[int]
    delay_secs: int
    def __init__(self, bus_id: _Optional[str] = ..., register_type: _Optional[int] = ..., address: _Optional[int] = ..., values: _Optional[_Iterable[int]] = ..., delay_secs: _Optional[int] = ...) -> None: ...

class scheduleWriteRegisterResponse(_message.Message):
    __slots__ = ("response_header", "bus_id", "delay_secs")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    DELAY_SECS_FIELD_NUMBER: _ClassVar[int]
    response_header: responseHeader
    bus_id: str
    delay_secs: int
    def __init__(self, response_header: _Optional[_Union[responseHeader, _Mapping]] = ..., bus_id: _Optional[str] = ..., delay_secs: _Optional[int] = ...) -> None: ...

class readRegisterSubscriptionRequest(_message.Message):
    __slots__ = ("bus_id", "modbus_id", "register_type", "address", "count", "poll_secs")
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    MODBUS_ID_FIELD_NUMBER: _ClassVar[int]
    REGISTER_TYPE_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    POLL_SECS_FIELD_NUMBER: _ClassVar[int]
    bus_id: str
    modbus_id: int
    register_type: int
    address: int
    count: int
    poll_secs: int
    def __init__(self, bus_id: _Optional[str] = ..., modbus_id: _Optional[int] = ..., register_type: _Optional[int] = ..., address: _Optional[int] = ..., count: _Optional[int] = ..., poll_secs: _Optional[int] = ...) -> None: ...

class readRegisterSubscriptionResponse(_message.Message):
    __slots__ = ("response_header", "bus_id", "response_code", "values")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    BUS_ID_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_CODE_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    response_header: responseHeader
    bus_id: str
    response_code: int
    values: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, response_header: _Optional[_Union[responseHeader, _Mapping]] = ..., bus_id: _Optional[str] = ..., response_code: _Optional[int] = ..., values: _Optional[_Iterable[int]] = ...) -> None: ...

class createServerRequest(_message.Message):
    __slots__ = ("type", "port", "host", "holding_registers", "input_registers", "coils", "discrete_inputs", "modbus_id")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    HOST_FIELD_NUMBER: _ClassVar[int]
    HOLDING_REGISTERS_FIELD_NUMBER: _ClassVar[int]
    INPUT_REGISTERS_FIELD_NUMBER: _ClassVar[int]
    COILS_FIELD_NUMBER: _ClassVar[int]
    DISCRETE_INPUTS_FIELD_NUMBER: _ClassVar[int]
    MODBUS_ID_FIELD_NUMBER: _ClassVar[int]
    type: str
    port: int
    host: str
    holding_registers: _containers.RepeatedScalarFieldContainer[int]
    input_registers: _containers.RepeatedScalarFieldContainer[int]
    coils: _containers.RepeatedScalarFieldContainer[bool]
    discrete_inputs: _containers.RepeatedScalarFieldContainer[bool]
    modbus_id: int
    def __init__(self, type: _Optional[str] = ..., port: _Optional[int] = ..., host: _Optional[str] = ..., holding_registers: _Optional[_Iterable[int]] = ..., input_registers: _Optional[_Iterable[int]] = ..., coils: _Optional[_Iterable[bool]] = ..., discrete_inputs: _Optional[_Iterable[bool]] = ..., modbus_id: _Optional[int] = ...) -> None: ...

class createServerResponse(_message.Message):
    __slots__ = ("response_header", "response")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    response_header: serverResponseHeader
    response: str
    def __init__(self, response_header: _Optional[_Union[serverResponseHeader, _Mapping]] = ..., response: _Optional[str] = ...) -> None: ...

class listServerRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class listServerResponse(_message.Message):
    __slots__ = ("response_header", "servers")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    SERVERS_FIELD_NUMBER: _ClassVar[int]
    response_header: serverResponseHeader
    servers: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, response_header: _Optional[_Union[serverResponseHeader, _Mapping]] = ..., servers: _Optional[_Iterable[str]] = ...) -> None: ...

class closeServerRequest(_message.Message):
    __slots__ = ("port",)
    PORT_FIELD_NUMBER: _ClassVar[int]
    port: int
    def __init__(self, port: _Optional[int] = ...) -> None: ...

class closeServerResponse(_message.Message):
    __slots__ = ("response_header", "port")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    response_header: serverResponseHeader
    port: int
    def __init__(self, response_header: _Optional[_Union[serverResponseHeader, _Mapping]] = ..., port: _Optional[int] = ...) -> None: ...

class setServerRegistersRequest(_message.Message):
    __slots__ = ("port", "function_code", "address", "values")
    PORT_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_CODE_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    port: int
    function_code: int
    address: _containers.RepeatedScalarFieldContainer[int]
    values: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, port: _Optional[int] = ..., function_code: _Optional[int] = ..., address: _Optional[_Iterable[int]] = ..., values: _Optional[_Iterable[int]] = ...) -> None: ...

class setServerRegistersResponse(_message.Message):
    __slots__ = ("response_header",)
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    response_header: serverResponseHeader
    def __init__(self, response_header: _Optional[_Union[serverResponseHeader, _Mapping]] = ...) -> None: ...

class getServerRegistersRequest(_message.Message):
    __slots__ = ("port", "function_code", "address", "count")
    PORT_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_CODE_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    port: int
    function_code: int
    address: int
    count: int
    def __init__(self, port: _Optional[int] = ..., function_code: _Optional[int] = ..., address: _Optional[int] = ..., count: _Optional[int] = ...) -> None: ...

class getServerRegistersResponse(_message.Message):
    __slots__ = ("response_header", "values")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    response_header: serverResponseHeader
    values: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, response_header: _Optional[_Union[serverResponseHeader, _Mapping]] = ..., values: _Optional[_Iterable[int]] = ...) -> None: ...

class scheduleServerRegistersRequest(_message.Message):
    __slots__ = ("port", "function_code", "address", "values", "delay_secs")
    PORT_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_CODE_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    DELAY_SECS_FIELD_NUMBER: _ClassVar[int]
    port: int
    function_code: int
    address: int
    values: _containers.RepeatedScalarFieldContainer[int]
    delay_secs: float
    def __init__(self, port: _Optional[int] = ..., function_code: _Optional[int] = ..., address: _Optional[int] = ..., values: _Optional[_Iterable[int]] = ..., delay_secs: _Optional[float] = ...) -> None: ...

class scheduleServerRegistersResponse(_message.Message):
    __slots__ = ("response_header",)
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    response_header: serverResponseHeader
    def __init__(self, response_header: _Optional[_Union[serverResponseHeader, _Mapping]] = ...) -> None: ...



import logging

from pydoover.cloud.processor import ProcessorBase


class HelloWorld(ProcessorBase):
    def setup(self): ...

    def process(self):
        logging.info("Hello World Started...")

        logging.debug("Triggerred by: %s", self.task_id)

        hello_world_channel = self.fetch_channel_named("josh-test")
        hello_world_channel.publish("Hello World1")

        logging.info("Hello World Finished")

    def close(self): ...


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    inst = HelloWorld(
        agent_id="agent_id",
        access_token="access_token",
        api_endpoint="https://my.doover.com",
        package_config={},
        msg_obj={},
        task_id="task_id",
        log_channel="log_channel",
        agent_settings={"deployment_config": {}},
    )
    inst.execute()



import inspect
import re

from typing import Any, Type

from .element import Element


NAME_VALIDATOR = re.compile(r"^[a-zA-Z0-9_-]+$")


class Container(Element):
    """Represents a container for UI elements, such as a submodule or application.

    This class can hold multiple child elements and provides methods to manage them.

    Generally, this class is not called by a user directly, but rather through the `Submodule` or `Application` classes.

    Parameters
    ----------
    name: str
        The name of the container, used for identification.
    display_name: str
        The display name of the container, used for user interface representation.
    children: list[Element]
        A list of child elements contained within this container.
    status_icon: str, optional
        An icon representing the status of the container, if applicable.
    auto_add_elements: bool
        If True, automatically adds all elements defined in the container to its children.

    Attributes
    ----------
    status_icon: str
        The icon representing the status of the container, if applicable.
    """

    type = "uiContainer"

    def __init__(
        self,
        name,
        display_name=None,
        children: list[Element] = None,
        status_icon: str = None,
        auto_add_elements: bool = True,
        **kwargs,
    ):
        super().__init__(name, display_name, **kwargs)

        self._default_position = 101
        self._max_position = self._default_position

        # A list of doover_ui_elements
        self._children = dict()
        self.add_children(*children or [])

        self.status_icon = status_icon

        self._register_interactions()
        if auto_add_elements:
            self.add_children(
                *[
                    e
                    for name, e in inspect.getmembers(
                        self, predicate=lambda e: isinstance(e, Element)
                    )
                ]
            )

    def _register_interactions(self):
        for name, func in inspect.getmembers(
            self, predicate=lambda f: inspect.ismethod(f) and hasattr(f, "_ui_type")
        ):
            item = func._ui_type(**func._ui_kwargs)
            item.callback = func
            setattr(self, func.__name__, item)

    def to_dict(self):
        result = super().to_dict()

        if self.status_icon is not None:
            result["statusIcon"] = self.status_icon

        result["children"] = {name: c.to_dict() for name, c in self._children.items()}
        return result

    def get_diff(
        self,
        other: dict[str, Any],
        remove: bool = True,
        retain_fields: list | None = None,
    ) -> dict[str, Any] | None:
        retain_fields = retain_fields or []
        res = super().get_diff(other, remove=remove, retain_fields=retain_fields) or {}
        # this will account for all the "normal" attributes, but not the children, since dicts aren't hashable
        # (ie. you can't do dict1 == dict2 to see if they're equal)
        other_children = other.get("children", {})
        this_children = {name: c for name, c in self._children.items()}

        children_diff = dict()
        if remove:
            children_diff.update(
                {k: None for k in other_children if k not in this_children}
            )  # to_remove

        for name, child in this_children.items():
            try:
                diff = child.get_diff(
                    other_children[name], remove=remove, retain_fields=retain_fields
                )
                if diff is not None:
                    children_diff[name] = diff
            except KeyError:
                children_diff[name] = child.to_dict()

        if children_diff:
            res["children"] = children_diff

        if len(res) == 0:
            return None

        return res

    @property
    def children(self) -> list[Element]:
        """Returns a list of child elements contained within this container."""
        return list(self._children.values())

    def set_children(self, children: list[Element]):
        """Sets the children of this container to a new list of elements.


        This will clear the existing children and replace them with the new ones.

        Warnings
        ---------
        This method should generally not be used by a user, instead use `add_children` to add elements to the container.
        Elements should generally never be removed from a container - instead, set them to `hidden=True`.


        Parameters
        ----------
        children: list[Element]
            A list of child elements to set as the new children of this container.
        """
        self._children.clear()
        self._max_position = self._default_position
        self.add_children(*children)

    def add_children(self, *children: Element):
        """Adds one or more child elements to this container.

        This method will automatically assign a position to each child if it does not already have one.

        Warnings
        --------
        You should generally only call this method once during setup of the container,
        when you generate all elements and add them at once. Old applications may call this multiple times during setup,
        but that is not the suggested best practice going forward.

        Parameters
        ----------
        *children
            Child elements to add to this container. They must be of type :class:`pydoover.ui.Element`
        """

        # if not hasattr(self, "_default_position"):
        #     self._default_position = 101
        # if not hasattr(self, "_max_position"):
        #     self._max_position = self._default_position

        for c in children:
            if not isinstance(c, Element):
                continue

            name = c.name.strip()
            if not NAME_VALIDATOR.match(name):
                raise RuntimeError(
                    f"Invalid name '{name}' for element '{c}'. Valid characters include letters, numbers, and underscores."
                )

            self._children[name] = c
            c.parent = self

            if not c.position:
                c.position = self._max_position
                self._max_position += 1

        return self

    def remove_children(self, *children: Element):
        """Removes one or more child elements from this container.

        Warnings
        --------
        Best practice prefers setting `hidden=True` on elements instead of removing them from the container.

        Parameters
        ----------
        *children
            Child elements to remove from this container. They must be of type :class:`pydoover.ui.Element`
        """
        for c in children:
            try:
                if c.name in self._children:
                    del self._children[c.name]
            except KeyError:
                pass

        ## for all self._children, call remove_children on them
        for c in self._children.values():
            if isinstance(c, Container):
                c.remove_children(*children)

    def clear_children(self):
        """Clears all child elements from this container.

        You probably don't want or need to call this method.
        """
        self._children.clear()

    def get_all_elements(
        self, type_filter: Type[Element] | None = None
    ) -> list[Element]:
        """Returns a list of all elements recursively contained within this container."""
        elements = []
        for element in self._children.values():
            if isinstance(element, Container):
                elements.extend(element.get_all_elements(type_filter))
            elif type_filter is None or isinstance(element, type_filter):
                elements.append(element)
        return elements

    def get_element(self, element_name: str) -> Element | None:
        """Retrieves a child element by its name from this container.

        This will recursively look through all children and their children to find the element.
        """
        try:
            return self._children[element_name]
        except KeyError:
            pass

        for element in self._children.values():
            if isinstance(element, Container):
                elem = element.get_element(element_name)
                if elem is not None:
                    return elem


class Submodule(Container):
    """Represents a submodule within a UI application, which can contain other elements and has a status.

    Submodules are useful for grouping logical components of an application together,
    but be careful not to overuse them as they can be burdensome on a user!

    Parameters
    ----------
    name: str
        The name of the submodule, used for identification.
    display_name: str
        The display name of the submodule, used for user interface representation.
    children: list[Element], optional
        A list of child elements contained within this submodule. Defaults to an empty list.
    status: str, optional
        A status string representing the current state of the submodule. Defaults to None.
    is_collapsed: bool, optional
        Whether the submodule is initially collapsed in the UI. Defaults to False.
    """

    type = "uiSubmodule"

    def __init__(
        self,
        name: str,
        display_name: str,
        children: list[Element] = None,
        status: str = None,
        is_collapsed: bool = False,
        **kwargs,
    ):
        super().__init__(name, display_name, children, **kwargs)

        self.status = status or kwargs.pop("status_string", None)
        self.collapsed = is_collapsed or kwargs.pop("collapsed", False)

    def to_dict(self):
        result = super().to_dict()
        if self.status is not None:
            result["statusString"] = self.status
        result["defaultOpen"] = not self.collapsed

        return result


class Application(Container):
    """Represents a UI application element.

    This is generally not invoked by the user, but is used to represent and store all UI elements for an application.

    Attributes
    ----------
    variant: str, optional
        The variant of the application, used to display applications differently.
        Defaults to `stacked`. Valid options are `stacked`, `submodule`.
    """

    type = "uiApplication"

    def __init__(self, *args, **kwargs):
        self.variant = kwargs.pop("variant", None)
        super().__init__(*args, **kwargs)

    def to_dict(self):
        result = super().to_dict()
        if self.variant is not None:
            result["variant"] = self.variant
        return result
    
    
class RemoteComponent(Container):
    """Represents a remote component in the UI.

    Parameters
    ----------
    name: str
        The name of the remote component.
    display_name: str
        The display name of the remote component.
    component_url: str
        The URL of the remote component.
    """

    type = "uiRemoteComponent"

    def __init__(
        self, name: str, 
        display_name: str, 
        component_url: str, 
        children: list[Element] = None, 
        **kwargs
    ):
        super().__init__(name, display_name, children, **kwargs)
        self.component_url = component_url
        self.kwargs = kwargs

    def to_dict(self):
        res = super().to_dict()
        res.update(self.kwargs)
        return res



import enum
import logging

from datetime import datetime
from typing import Optional, Any

from .misc import Colour


class NotSet:
    pass


log = logging.getLogger(__name__)


class Element:
    """Base class for UI elements.

    Attributes
    ----------
    name: str
        The name of the UI element.
    display_name: str, optional
        The display name of the UI element.
    is_available: bool, optional
        Indicates if the UI element is available.
    help_str: str, optional
        A help string for the UI element.
    verbose_str: str, optional
        A verbose string for the UI element.
    show_activity: bool
        Whether to show activity for the UI element.
    form: str, optional
        The form associated with the UI element.
    graphic: str, optional
        The graphic associated with the UI element.
    layout: str, optional
        The layout associated with the UI element.
    component_url: str, optional
        The URL for a remote component associated with the UI element.
    position: int, optional
        The position of the UI element in the UI.
    conditions: dict[str, Any], optional
        Conditions under which the UI element is displayed.
    hidden: bool
        Whether the UI element is hidden.
    """

    type = "uiElement"

    def __init__(
        self,
        name: str,
        display_name: str | None = None,
        is_available: bool = None,  # not sure of type
        help_str: str = None,
        verbose_str: str = None,
        show_activity: bool = True,
        form: str = None,
        graphic: str = None,  # not sure of type
        layout: str = None,  # not sure of type
        component_url: str = None,  # not sure of type
        position: Optional[int] = None,  # 100,
        conditions: Optional[dict] = None,
        hidden: bool = False,
        **kwargs,
    ):
        self.name = name
        self.display_name = display_name or kwargs.get(
            "display_str", None
        )  # backwards compatibility
        self.is_available = is_available
        self.help_str = help_str
        self.verbose_str = verbose_str
        self.show_activity = show_activity
        self.form = form
        self.graphic = graphic
        self.layout = layout
        self.component_url = component_url
        self.position = position
        self.conditions = conditions
        self.hidden = hidden

        for k, v in kwargs.items():
            log.debug(f"Setting kwarg {k} to {v}")
            setattr(self, k, v)

        self._retain_fields = []

    def to_dict(self):
        """Convert the element to a dictionary representation.

        Returns
        -------
        dict[str, Any]
            A dictionary representation of the UI element. This is used to serialize the element for the site API.
        """
        to_return = {
            "name": self.name,
            "type": self.type,
            "displayString": self.display_name,
            "isAvailable": self.is_available,
            "helpString": self.help_str,
            "verboseString": self.verbose_str,
            "showActivity": self.show_activity,
            "form": self.form,
            "graphic": self.graphic,
            "layout": self.layout,
            "componentUrl": self.component_url,
            "position": self.position,
            "conditions": self.conditions,
            "hidden": self.hidden,
        }
        # filter out any null values
        return {k: v for k, v in to_return.items() if v is not None}

    def get_diff(
        self,
        other: dict[str, Any],
        remove: bool = True,
        retain_fields: list[str] | None = None,
    ) -> dict[str, Any] | None:
        """Get the difference between this element and another dictionary representation of an element.

        Parameters
        ----------
        other: dict[str, Any]
            The other element to compare against.
        remove: bool
            If True, fields that are not present in this element but are in the other will be set to None.
        retain_fields: list[str], optional
            A list of fields to retain in the difference, even if they are not different.

        Returns
        -------
        dict[str, Any], optional
            A dictionary containing the differences between this element and the other. If there are no differences, returns None.
        """

        this = self.to_dict()

        retain_fields = retain_fields or []
        to_retain = list(set(retain_fields) | set(self._retain_fields))
        # if this == other:
        #     return None

        result = {k: v for k, v in this.items() if other.get(k) != v or k in to_retain}
        if remove:
            result.update(
                **{k: None for k in other if k not in this and k not in to_retain}
            )  # to_remove
        if len(result) == 0:
            return None

        return result

    ## A stub for the method that will be called when the UI state is updated.
    # The element can choose to update its internal state based on the previous state and the new state.
    def recv_ui_state_update(self, state: dict[str, Any]) -> None:
        pass


class ConnectionType(enum.Enum):
    constant = "constant"
    periodic = "periodic"
    other = "other"


class ConnectionInfo(Element):
    """Connection Info

    Parameters
    ----------
    name: str
    connection_type: ConnectionType

    connection_period: int
        The expected time between connection events (seconds) - only applicable for "periodic"
    next_connection: int
        Expected time of next connection (seconds after shutdown)
    offline_after: int
        Show as offline if disconnected for more than x secs

    """

    type = "uiConnectionInfo"

    def __init__(
        self,
        name: str = "connectionInfo",
        connection_type: ConnectionType = ConnectionType.constant,
        connection_period: int = None,
        next_connection: int = None,
        offline_after: int = None,
        allowed_misses: int = None,
        **kwargs,
    ):
        super().__init__(name, None, **kwargs)
        self.name = name
        self.connection_type = connection_type
        self.connection_period = connection_period
        self.next_connection = next_connection
        self.offline_after = offline_after
        self.allowed_misses = allowed_misses

        if self.connection_type is not ConnectionType.periodic and (
            self.connection_period is not None
            or self.next_connection is not None
            or self.allowed_misses is not None
        ):
            raise RuntimeError(
                "connection_type must be periodic to set connection_period, next_connection or offline_after"
            )

    def to_dict(self):
        result = {
            "name": self.name,
            "type": self.type,
            "connectionType": self.connection_type.value,
        }

        if self.connection_period is not None:
            result["connectionPeriod"] = self.connection_period
        if self.next_connection is not None:
            result["nextConnection"] = self.next_connection
        if self.offline_after is not None:
            result["offlineAfter"] = self.offline_after
        if self.allowed_misses is not None:
            result["allowedMisses"] = self.allowed_misses

        return result


class AlertStream(Element):
    """Represents an Alert Stream UI Element

    .. note::

        This is a special element that is used to display the "Notifications" banner in the UI.
        If any installed app includes this element, it will be shown.
        Do not change the name of this element, doing so will lead to confusion as
        it is manually set to listen to the "significantEvent" channel in the UI.

    Parameters
    ----------
    name: str
        The name of the alert stream.
        This defaults to "significantEvent", but is currently unused in the UI.
    display_name: str, optional
        The display name of the alert stream. This is not used in the UI.
    """

    type = "uiAlertStream"

    def __init__(
        self,
        name: str = "significantEvent",
        display_name: str = "placeholder",
        **kwargs,
    ):
        super().__init__(name, display_name, is_available=None, help_str=None, **kwargs)


class Multiplot(Element):
    """Represents a MultiPlot UI Element.

    Parameters
    ----------
    name: str
        The name of the multiplot.
    display_name: str
        The display name of the multiplot.
    series: list[str]
        A list of series names to be displayed in the multiplot.
    series_colours: list[Colour], optional
        A list of colours for each series in the multiplot.
    series_active: list[bool], optional
        A list indicating whether each series is active or not.
    earliest_data_time: datetime, optional
        The earliest time for which data is available in the multiplot.
    title: str, optional
        The title of the multiplot.
    """

    type = "uiMultiPlot"

    def __init__(
        self,
        name: str,
        display_name: str,
        series: list[str],
        series_colours: Optional[list[Colour]] = None,
        series_active: Optional[list[bool]] = None,
        earliest_data_time: Optional[datetime] = None,
        title: Optional[str] = None,
        shared_axis: Optional[list[bool]] = None,
        step_labels: Optional[list[str]] = None,
        step_padding: Optional[list[int]] = None,
        default_zoom: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(name, display_name, **kwargs)

        self.series = series
        self.series_colours = series_colours
        self.series_active = series_active
        self.earliest_data_time = earliest_data_time
        self.title = title
        self.shared_axis = shared_axis
        self.step_labels = step_labels
        self.step_padding = step_padding
        self.default_zoom = default_zoom

    def to_dict(self):
        result = super().to_dict()
        result["series"] = self.series
        result["colours"] = self.series_colours

        if self.series_active is not None:
            result["activeSeries"] = self.series_active
        if self.shared_axis is not None:
            result["sharedAxis"] = self.shared_axis
        if self.step_labels is not None:
            result["stepLabels"] = self.step_labels
        if self.step_padding is not None:
            result["stepPadding"] = self.step_padding
        if self.default_zoom is not None:
            result["defaultZoom"] = self.default_zoom
        if self.title is not None:
            result["title"] = self.title

        if self.earliest_data_time is not None:
            if isinstance(self.earliest_data_time, datetime):
                result["earliestDataDate"] = int(self.earliest_data_time.timestamp())
            else:
                result["earliestDataDate"] = self.earliest_data_time

        return result




#!/usr/bin/env python3
import os
from datetime import datetime

from .data import ReportGeneratorDooverDataMixin


class ReportGeneratorBase:
    """
    Report generator that creates reports for given agents within a specified time period.

    Attributes:
        tmp_workspace (str): Temporary workspace path.
        access_token (str): API access token.
        agent_ids (list): List of agent IDs.
        agent_display_names (list): List of agent display names.
        period_from (datetime): Start of the period (in UTC).
        period_to (datetime): End of the period (in UTC).
        for_timezone (pytz.timezone): Timezone for localizing the report.
        logging_function (callable): Optional function for logging.
        progress_update_function (callable): Optional function to update progress.
        api_endpoint (str): API endpoint.
        report_name (str): Name of the report.
        test_mode (bool): Flag for test mode.
    """

    def __init__(
        self,
        tmp_workspace: str,
        access_token: str,
        agent_ids: list,
        agent_display_names: list,
        period_from_utc: datetime,
        period_to_utc: datetime,
        for_timezone,
        logging_function=None,
        progress_update_function=None,
        api_endpoint: str = None,
        report_name: str = None,
        test_mode: bool = False,
    ):
        self.tmp_workspace = tmp_workspace
        self.access_token = access_token
        self.agent_ids = agent_ids
        self.agent_display_names = agent_display_names
        self.period_from = period_from_utc  # UTC datetime
        self.period_to = period_to_utc  # UTC datetime
        self.for_timezone = for_timezone
        self.logging_function = logging_function
        self.progress_update_function = progress_update_function
        self.api_endpoint = api_endpoint
        self.report_name = report_name
        self.test_mode = test_mode

        self._localize_period()

        if self.logging_function:
            self.logging_function("Report Generator Initialized")
            self.logging_function(f"Agents = {self.agent_ids}")
            self.logging_function(f"Start Period = {self.period_from}")
            self.logging_function(f"End Period = {self.period_to}")
            self.logging_function(f"For Timezone = {self.for_timezone}")

    def update_progress(self, progress: float, *args, **kwargs) -> None:
        """
        Update the progress of the report generation.

        Args:
            progress (float): Progress value between 0 and 1.
        """
        if self.progress_update_function:
            self.progress_update_function(progress, *args, **kwargs)

    def _localize_period(self):
        """
        Localize the UTC period to the desired timezone.
        Returns a tuple (localized_period_from, localized_period_to).
        """
        period_from_naive = self.period_from.replace(tzinfo=None)
        period_to_naive = self.period_to.replace(tzinfo=None)

        if self.for_timezone:
            localized_period_from = self.for_timezone.fromutc(period_from_naive)
            localized_period_to = self.for_timezone.fromutc(period_to_naive)
        else:
            localized_period_from = period_from_naive
            localized_period_to = period_to_naive

        self.localized_period_from = localized_period_from
        self.localized_period_to = localized_period_to

        return localized_period_from, localized_period_to

    def assess_progress(self, progress: float, agent_id: str = None) -> None:
        """
        Assess the progress of the report generation.

        Args:
            progress (float): Progress value between 0 and 1.
            agent_id (str): The optional id of the agent currently in question
        """
        if not agent_id:
            self.update_progress(progress)
            return

        ## Update as if doing multiple reports for each individual agent
        try:
            ind = self.agent_ids.index(agent_id)
        except ValueError:
            if self.logging_function:
                self.logging_function(
                    f"Agent ID {agent_id} not found in the agent list."
                )
            self.update_progress(progress)
            return

        full_len = len(self.agent_ids)
        total_progress = (ind / full_len) + (progress / full_len)

        self.update_progress(total_progress)

    ## Abstract methods to optionally define
    # def generate_one(self, agent_id: str) -> None:
    #     """
    #     Generate a report for a single agent.
    #     """
    #     pass

    def generate(self) -> None:
        """
        Generate a report
        This is designed to either be called directly or to be overridden by a subclass.

        Generally speaking, users should either override this method or the generate_for_one_agent method.
        """
        if self.logging_function:
            self.logging_function("Generating report...")

        if hasattr(self, "generate_one"):
            for agent_id in self.agent_ids:
                self.assess_progress(0.01, agent_id)
                self.generate_one(agent_id)
                self.assess_progress(1.0, agent_id)
        else:
            if self.logging_function:
                self.logging_function("No report generation method defined.")

    def get_agent_display_name(self, agent_id: str) -> str:
        """
        Get the display name for an agent based on its ID.
        """
        try:
            ind = self.agent_ids.index(agent_id)
        except ValueError:
            return None

        return self.agent_display_names[ind]

    def add_to_log(self, msg):
        """
        Log a message via the provided logging_function (or print if not provided).
        """
        if self.logging_function is not None:
            self.logging_function(msg)
        else:
            print(msg)

    def get_ws_path(self, *args) -> str:
        """
        Get the path to a file in the temporary workspace.
        """
        if self.tmp_workspace is None:
            return os.path.join(*args)
        return os.path.join(self.tmp_workspace, *args)

    def get_file_outputs(self) -> list:
        """
        Returns a list of expected output filenames for each agent.
        """
        return [f"{name}.xlsx" for name in self.agent_display_names]

    def get_output_message(self) -> str:
        """
        Returns the output message after report generation.
        """
        return "Report generated"


class ReportGenerator(ReportGeneratorBase, ReportGeneratorDooverDataMixin):
    pass


generator = ReportGenerator



from pathlib import Path

from pydoover import config


class SampleConfig(config.ApplicationConfig):
    def __init__(self):
        # these 2 are device specific, and inherit from the device-set variables.
        # However, the user can override them if they wish.

        self.num_di = config.Integer(
            "Digital Input Count",
            default=config.Variable("device", "digitalInputCount"),
            min_val=0,
        )
        self.num_do = config.Integer(
            "Digital Output Count",
            default=config.Variable("device", "digitalOutputCount"),
            min_val=0,
        )

        self.outputs_enabled = config.Boolean("Digital Outputs Enabled", default=True)
        self.funny_message = config.String(
            "A Funny Message"
        )  # this will be required as no default given.


if __name__ == "__main__":
    c = SampleConfig()
    c.export(Path("app_config.json"))



# class TestClass:
#     def __init__(self):
#         ## A list of return values simulating a battery voltage with an intermittent drop
#         self.values = (
#             [
#                 12.0,
#                 11.5,
#                 11.0,
#                 10.0,
#                 10.5,
#                 11.0,
#                 11.5,
#                 5,
#                 12.0,
#                 12.5,
#                 13.0,
#                 13.5,
#                 14.0,
#                 14.5,
#                 15.0,
#             ]
#             + [0] * 30
#             + [12] * 25
#             + [None] * 3
#             + [8] * 10
#         )
#
#         ## An alternate list the same as the batteries but negative
#         # self.values = [-v for v in self.values if v is not None]
#
#         ## Another alternate list, but of a much smaller magnitude
#         # self.values = [v/100 for v in self.values if v is not None]
#
#         ## another alternate list, but with a much larger magnitude
#         self.values_2 = [v * 100 for v in self.values if v is not None]
#
#         # ## A list of return values simulating a gas ppm sensor with an intermittent spike
#         # self.values = [
#         #     5000, 4900, 4800, 4700, 4850, 5000, 5150, 5300, 5250, 5200, 5150, 5100, 5050, 5000, 4950, 4900,
#         #     20000, 19500, 19000, 500, 600, 550, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500,
#         #     5000, 4900, 4800, 4700, 4850, 5000, 5150, 5300, 5250, 5200, 5150, 5100, 5050, 5000, 4950, 4900,
#         #     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
#         # ]
#
#         self.counter = 0
#
#     ## define a generator function that returns the next sensor value
#     def _next(self):
#         # increment the counter
#         self.counter += 1
#
#     def is_running(self):
#         return self.counter < len(self.values)
#
#     def _raw_sensor_value(self):
#         if self.counter < len(self.values):
#             return self.values[self.counter]
#         return None
#
#     def _raw_sensor_value_2(self):
#         if self.counter < len(self.values):
#             return self.values_2[self.counter]
#         return None
#
#     ## Can optionally provide initial_estimate, initial_error_estimate, and process_variance, among others here also
#     @apply_kalman_filter()
#     def sensor_value(self, kf_process_variance=None):
#         return self._raw_sensor_value()
#
#     @apply_kalman_filter()
#     def sensor_value_2(self):
#         return self._raw_sensor_value_2()
#
#
# logging.basicConfig(level=logging.DEBUG)
#
# test = TestClass()
# # logging.info(f"Raw values: {test.values}")
#
# while test.is_running():
#     filtered_value = test.sensor_value(kf_process_variance=0.001)
#     # filtered_value_2 = test.sensor_value_2()
#     # Can also provide measurement_variance and dt as keyword arguments. e.g.
#     # filtered_value = test.sensor_value(kf_measurement_variance=0.5, kf_dt=1)
#
#     # logging.info(f"Raw value: {test._raw_sensor_value()}, Filtered value: {filtered_value}")
#     test._next()
#     time.sleep(1)



import random
import time

from pydoover.docker import DockerApplication, run_app
from pydoover import ui

# UI Will look like this

# Variable : Is Working : Bool
# Variable : Uptime : Int
# Parameter : Test Message
# Variable : Test Output
# Action : Send this text as an alert
# Submodule :
#      Variable : Battery Voltage
#      Parameter : Low Battery Voltage Alert
#            Once below this setpoint, send a text and show a warning
#      StateCommand : Charge Battery Mode
#           - Charge
#           - Discharge
#           - Idle


class HelloWorld(DockerApplication):
    started: time.time
    is_working: ui.BooleanVariable
    send_alert: ui.Action
    on_text_parameter_change: ui.TextParameter

    def setup(self):
        super().setup()
        self.get_ui_manager()

        self.started = time.time()
        include_uptime = True

        # Define the UI
        self.is_working = ui.BooleanVariable("is_working", "We Working?")
        ui_elems = (
            self.is_working,
            ui.DateTimeVariable("uptime", "Started") if include_uptime else None,
            self.send_alert,
            # ui.TextParameter("test_message", "Put in a message", callback=self.on_text_parameter_change),
            self.on_text_parameter_change,
            ui.TextVariable("test_output", "This is message we got"),
            ui.Submodule(
                "battery",
                "Battery Module",
                children=[
                    ui.NumericVariable(
                        "voltage",
                        "Battery Voltage",
                        precision=2,
                        ranges=[
                            ui.Range("Low", 0, 10, ui.Colour.red),
                            ui.Range("Normal", 10, 20, ui.Colour.green),
                            ui.Range("High", 20, 30, ui.Colour.blue),
                        ],
                    ),
                    ui.NumericParameter("low_voltage_alert", "Low Voltage Alert"),
                    ui.StateCommand(
                        "charge_mode",
                        "Charge Mode",
                        callback=self.on_state_command,
                        user_options=[
                            ui.Option("charge", "Charge"),
                            ui.Option("discharge", "Discharge"),
                            ui.Option("idle", "Idle"),
                        ],
                    ),
                ],
            ),
        )

        self.ui_manager.add_children(*ui_elems)

    def main_loop(self):
        super().main_loop()
        self.is_working.current_value = True
        self.ui_manager.update_variable("voltage", random.randint(900, 2100) / 100)
        self.ui_manager.update_variable("uptime", time.time() - self.started)

    @ui.action("send_alert", "Send message as alert", position=1)
    def send_alert(self, new_value):
        output = self.ui_manager.get_element("test_output").current_value
        self.log(f"Sending alert: {output}")
        self.send_notification(output, record_activity=True)
        self.send_alert.coerce(None)

    @ui.text_parameter("test_message", "Put in a message")
    def on_text_parameter_change(self, new_value):
        self.log("New value for test message : " + new_value)

        # Set the value as an output to the corresponding variable is this case
        self.get_ui_manager().update_variable("test_output", new_value)

    def on_state_command(self, new_value):
        self.log("New value for state command: " + new_value)


if __name__ == "__main__":
    new_app = HelloWorld()
    run_app(new_app)



import asyncio
import json
import logging

from collections.abc import MutableMapping

from functools import wraps
from typing import Any

log = logging.getLogger(__name__)


def map_reading(in_val, output_values, raw_readings=[4, 20], ignore_below=3):
    """Map a reading to a value in a range"""
    if in_val < ignore_below:
        return None

    ## Choose the value set to map between
    lower_val_ind = 0
    found = False
    for i in range(0, len(raw_readings)):
        if in_val <= raw_readings[i]:
            lower_val_ind = i - 1
            found = True
            break

    if not found:
        lower_val_ind = len(raw_readings) - 2

    # Figure out how 'wide' each range is
    inSpan = raw_readings[lower_val_ind + 1] - raw_readings[lower_val_ind]
    outSpan = output_values[lower_val_ind + 1] - output_values[lower_val_ind]

    # Convert the left range into a 0-1 range (float)
    valueScaled = float(in_val - raw_readings[lower_val_ind]) / float(inSpan)

    # Convert the 0-1 range into a value in the right range.
    return output_values[lower_val_ind] + (valueScaled * outSpan)


def find_object_with_key(obj: dict[Any, Any], key_to_find: str) -> Any | None:
    """Iteratively searches through a dictionary (JSON object) and returns the value for the specified key.

    Parameters
    ----------
    obj : dict
        The JSON object (dictionary) to search through.
    key_to_find : str
        The key to search for.

    Returns
    -------
    Any
        The object containing the key, or None if the key is not found.
    """
    stack = [obj]

    while stack:
        current = stack.pop()

        if isinstance(current, dict):
            if key_to_find in current:
                return current[key_to_find]

            for key in current:
                stack.append(current[key])

    return None


def find_path_to_key(obj: dict[Any, Any], key_to_find: str) -> str | None:
    """Iteratively searches through a dictionary (JSON object) and returns the path to the specified key.

    Parameters
    ----------
    obj : dict
        The JSON object (dictionary) to search through.
    key_to_find : str
        The key to search for.

    Returns
    -------
    str, optional
        The path to the key, or None if the key is not found.

    """

    stack = [{"current": obj, "path": ""}]

    while stack:
        current_entry = stack.pop()
        current = current_entry["current"]
        path = current_entry["path"]

        if isinstance(current, dict):
            if key_to_find in current:
                return f"{path}.{key_to_find}" if path else key_to_find

            for key in current:
                new_path = f"{path}.{key}" if path else key
                stack.append({"current": current[key], "path": new_path})

    return None


def get_is_async(is_async: bool = None):
    if is_async is not None:
        return is_async

    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return False
    else:
        return True


def maybe_async():
    """Wrapper to allow both a sync and async variation on the same function to provide a unified interface to the user.

    This is useful when writing library functions for both a sync and async context.

    It assumes you have a variation of your function with the same signature suffixed with `_async`.

    Examples
    --------

    A simple example::

        class MyClass:
            @maybe_async()
            def my_function(self, value: str):
                print("This is the sync version of my_function")
                return "sync result"

            async def my_function_async(self, value: str):
                print("This is the async version of my_function")
                return "async result"



    If the user was running an asynchronous main loop, `my_function` would be invoked as follows::

        class MyApp(Application):
            async def main_loop(self):
                result = await obj.my_function("test")
                print(result)  # This would print "async result"


    However, if the main loop was synchronous, it would invoke the sync version of the function::

        class MyApp(Application):
            def main_loop(self):
                result = self.my_function("test")
                print(result)  # This would print "sync result"

    """

    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            # args[0] is self
            is_async_context = getattr(args[0], "_is_async", False)
            # allow them to specify if they want to run the sync version of the function (default very much not)
            force_sync = kwargs.pop("run_sync", False)

            if is_async_context is True and force_sync is False:
                # we're in an async context, check if we have an async variety of the function to run...
                try:
                    alternative = getattr(args[0], f"{func.__name__}_async")
                    return alternative(*args[1:], **kwargs)
                except AttributeError:
                    # we don't have a corresponding async method, just use the sync one.
                    pass
            return func(*args, **kwargs)

        return inner

    return wrapper


def wrap_try_except(func, *args, **kwargs):
    """Wrapper function to catch exceptions and log them. This does not propagate the exception."""
    try:
        return func(*args, **kwargs)
    except Exception as e:
        log.exception(f"Error in {func.__name__}: {e}", exc_info=e)


async def wrap_try_except_async(func, *args, **kwargs):
    """Wrapper function to catch exceptions in an async function and log them. This does not propagate the exception."""
    try:
        return await func(*args, **kwargs)
    except Exception as e:
        log.exception(f"Error in {func.__name__}: {e}", exc_info=e)


async def call_maybe_async(
    func, *args, as_task: bool = False, in_executor: bool = True, **kwargs
):
    """Helper function to call a function that may be either synchronous or asynchronous.

    Parameters
    ----------
    func : callable
        The function to call, which may be synchronous or asynchronous.

    *args
        Arguments to pass to the function.

    as_task : bool
        If True, the function will be called as an asyncio task. Default is False.

    in_executor : bool
        If True, the function will be run in an executor if it is not a coroutine function. Default is True.

    **kwargs
        Any kwargs to pass to the function.
    """
    # print(f"call_maybe_async: func={func}, as_task={as_task}, in_executor={in_executor is True and not asyncio.iscoroutinefunction(func)}")
    if asyncio.iscoroutinefunction(func):
        # coro = wrap_try_except_async(func, *args, **kwargs)
        coro = func(*args, **kwargs)
        if as_task:
            # assign it to a variable for weak ref
            task = asyncio.create_task(coro)
            return task
        else:
            return await coro
    elif in_executor:
        loop = asyncio.get_running_loop()
        # run_in_executor doesn't support kwargs
        if kwargs:
            log.warning("kwargs are not supported when calling via executor")

        # this is a little bit of a hack, but essentially we're creating an async function that
        # is called with await func to allow for both running as a task and in an executor.
        # future = loop.run_in_executor(None, wrap_try_except, func, *args)
        future = loop.run_in_executor(None, func, *args, **kwargs)
        if as_task:
            return future

        return await future
    else:
        # return wrap_try_except(func, catch_exception, *args, **kwargs)
        return func(*args, **kwargs)


def on_change(callback, name=None):
    """A decorator that triggers a callback when the output of the decorated function changes.

    The callback is called with four arguments:
      - new_result: The new output of the function.
      - old_result: The previous output (or None if this is the first call).
      - is_first: A boolean indicating if this is the first time the function has returned a value.
      - change_detector_name: The optional name identifying this change detector.

    :param callback: The callback function to trigger, or a string indicating the name of an instance attribute.
    :param name: An optional name for this change_detector_name.

    Examples
    ---------

    A simple usage example::


        class MyClass:
            def __init__(self):
                self.last = None

            def my_callback(self, new_result, old_result, is_first, change_detector_name):
                if is_first:
                    print(f"{change_detector_name} has returned a value for the first time: {new_result}")
                else:
                    print(f"{change_detector_name} has changed from {old_result} to {new_result}")

            @on_change("my_callback", name="my_function")
            def my_function(self):
                import random
                return random.randint(0, 100)

    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            detector_name = name or func.__name__

            # Check if we're dealing with an instance method.
            if args and hasattr(args[0], "__dict__"):
                instance = args[0]
                state_attr = f"__on_change_state_{detector_name}"
                # Retrieve state from the instance, or initialize it if not present.
                state = getattr(
                    instance, state_attr, {"last_result": None, "has_value": False}
                )
            else:
                # For non-instance functions, fallback to closure state.
                # (You could also raise an error if you only expect instance methods.)
                state = wrapper.__dict__.setdefault(
                    "state", {"last_result": None, "has_value": False}
                )

            result = func(*args, **kwargs)
            is_first = not state["has_value"]

            # Trigger the callback if it's the first call or if the result has changed.
            if is_first or result != state["last_result"]:
                # Determine which callback to use.
                if isinstance(callback, str):
                    if not args:
                        raise ValueError(
                            "Expected a 'self' argument when using a string callback."
                        )
                    cb = getattr(args[0], callback, None)
                    if cb is None or not callable(cb):
                        raise ValueError(
                            f"Attribute '{callback}' is not a callable of the instance."
                        )
                else:
                    cb = callback

                cb(
                    result,
                    state["last_result"] if state["has_value"] else None,
                    is_first,
                    detector_name,
                )
                state["last_result"] = result
                state["has_value"] = True

                # Save the updated state back to the instance (or closure, as applicable).
                if args and hasattr(args[0], "__dict__"):
                    setattr(instance, state_attr, state)
                else:
                    wrapper.__dict__["state"] = state

            return result

        return wrapper

    return decorator


class CaseInsensitiveDict(MutableMapping):
    def __init__(self, data=None, **kwargs):
        self._store = dict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def copy(self):
        return CaseInsensitiveDict(self._store)

    def to_dict(self):
        return {
            k: v.to_dict() if isinstance(v, CaseInsensitiveDict) else v
            for k, v in self.items()
        }

    @classmethod
    def from_dict(cls, data):
        t = {
            k: CaseInsensitiveDict.from_dict(v) if isinstance(v, dict) else v
            for k, v in data.items()
        }
        return cls(t)

    def __len__(self):
        return len(self._store)

    def __iter__(self):
        return iter(self._store)

    def __setitem__(self, key, value):
        self._store[key.lower()] = value

    def __getitem__(self, key):
        return self._store[key.lower()]

    def __delitem__(self, key):
        del self._store[key.lower()]

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self._store)


class CaseInsensitiveDictEncoder(json.JSONEncoder):
    # might not need this
    def default(self, obj):
        if isinstance(obj, CaseInsensitiveDict):
            return obj.to_dict()
        # Let the base class default method raise the TypeError
        return super().default(obj)


class LogFormatter(logging.Formatter):
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = (
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    )

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def setup_logging(debug: bool, formatter: logging.Formatter = None, filters: logging.Filter | list[logging.Filter] = None):
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    formatter = formatter or LogFormatter()

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    if filters:
        if isinstance(filters, logging.Filter):
            filters = [filters]
        for filter in filters:
            handler.addFilter(filter)

    logging.getLogger().handlers.clear()
    logging.getLogger().addHandler(handler)



import logging

from collections import namedtuple
from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Optional, TypeVar
from urllib.parse import quote, urlencode

import requests

from .message import Message
from .agent import Agent
from .application import Application
from .channel import Channel, Processor, Task
from .config import ConfigManager
from .exceptions import NotFound, Forbidden, HTTPException


log = logging.getLogger(__name__)
AccessToken = namedtuple("AccessToken", ["token", "expires_at"], defaults=(None,))
T = TypeVar("T", bound=Channel)


class Route:
    def __init__(self, method, route, *args, **kwargs):
        self.method = method

        self.url = route
        if args:
            self.url = route.format(*[quote(a) for a in args])

        if kwargs:
            self.url = f"{self.url}?{urlencode(kwargs)}"


class Client:
    """API Client for Doover Cloud"""

    def __init__(
        self,
        username: str = None,
        password: str = None,
        token: str = None,
        token_expires: datetime = None,
        base_url: str = "https://my.doover.dev",
        agent_id: str = None,
        verify: bool = True,
        login_callback: Callable = None,
        config_profile: str = "default",
        debug: bool = False,
    ):
        self.access_token = None
        if token:
            self.access_token = AccessToken(token, token_expires)

        self.agent_id = agent_id
        self.login_callback = login_callback

        self.username = username
        self.password = password

        self.verify = verify
        self.base_url = base_url
        self.session = requests.Session()

        self.request_retries = 1
        self.request_timeout = 59

        if debug:
            logging.basicConfig(level=logging.DEBUG)

        if not ((username and password) or token):
            self.config_manager = ConfigManager(config_profile)
            self.config_manager.read()

            config = self.config_manager.current
            if not config:
                raise RuntimeError(
                    f"No configuration found for profile {self.config_manager.current_profile}. "
                    f"Please specify a profile with the `config_profile` parameter, "
                    f"manually set a token or `doover login`"
                )

            self.agent_id = config.agent_id
            self.username = config.username
            self.password = config.password
            self.access_token = AccessToken(config.token, config.token_expires)
            self.base_url = config.base_url

        if self.access_token:
            self.update_headers()

    def has_persistent_connection(self):
        return False

    def update_headers(self):
        self.session.headers.update(
            {"Authorization": f"Token {self.access_token.token}"}
        )
        self.session.verify = self.verify

    def request(self, route: Route, **kwargs):
        # default is access token to not expire
        if self.access_token.expires_at and self.access_token.expires_at < datetime.now(
            timezone.utc
        ):
            logging.info("Token expired, attempting to refresh token.")
            self.login()

        url = self.base_url + route.url

        attempt_counter = 0
        retries = self.request_retries if route.method == "GET" else 0

        while attempt_counter <= retries:
            attempt_counter += 1

            log.debug(f"Making {route.method} request to {url} with kwargs {kwargs}")

            try:
                resp = self.session.request(
                    route.method, url, timeout=self.request_timeout, **kwargs
                )
            except requests.exceptions.Timeout:
                log.info(f"Request to {url} timed out.")
                if attempt_counter > retries:
                    raise HTTPException(f"Request timed out. {url}")
                continue

            if 200 <= resp.status_code < 300:
                # if we get a 200, we're good to go
                # this also accounts for any 200-range code.
                break
            elif resp.status_code == 403:
                raise Forbidden(f"Access denied. {url}")
            elif resp.status_code == 404:
                raise NotFound(f"Resource not found. {url}")

            log.info(
                f"Failed to make request to {url}. Status code: {resp.status_code}, message: {resp.text}"
            )
            if attempt_counter > retries:
                raise HTTPException(resp.text)

        try:
            data = resp.json()
        except ValueError:
            data = resp.text

        log.debug(f"{url} has received {data}")
        return data

    def _get_agent_raw(self, agent_id: str) -> dict[str, Any]:
        return self.request(Route("GET", "/ch/v1/agent/{}/", agent_id))

    def _get_agent_list_raw(self) -> list[dict[str, Any]]:
        return self.request(Route("GET", "/ch/v1/list_agents/"))

    def get_agent(self, agent_id: str) -> Agent:
        """Fetches an agent by its key.

        Parameters
        ----------
        agent_id : str
            The unique identifier for the agent.

        Returns
        -------
        Agent
            The agent object.

        Raises
        ------
        NotFound
            You don't have permission to access this agent.
        """
        data = self._get_agent_raw(agent_id)
        return data and Agent(client=self, data=data)

    def get_agent_list(self) -> list[Agent]:
        """Fetches a list of all agents you have access to.

        Returns
        -------
        list[Agent]
            A list of Agents that you have access to.
        """
        data = self._get_agent_list_raw()
        if "agents" not in data:
            return []
        return [Agent(client=self, data=d) for d in data["agents"]]

    def _parse_channel(self, data) -> T:
        if data["name"].startswith("!"):
            return Task(client=self, data=data)
        elif data["name"].startswith("#"):
            return Processor(client=self, data=data)
        else:
            return Channel(client=self, data=data)

    def _get_channel_raw(self, channel_id: str) -> dict[str, Any]:
        return self.request(Route("GET", "/ch/v1/channel/{}/", channel_id))

    def get_channel(self, channel_id: str) -> Optional[T]:
        """Fetch a channel by its key.

        Parameters
        ----------
        channel_id : str
            The unique identifier for the channel.

        Returns
        -------
        Channel, Processor or Task
            The channel object, which can be a Channel, Processor or Task depending on the type of the channel.
        """
        data = self._get_channel_raw(channel_id)
        return data and self._parse_channel(data)

    def _get_channel_named_raw(
        self, channel_name: str, agent_id: str
    ) -> dict[str, Any]:
        return self.request(Route("GET", "/ch/v1/agent/{}/{}/", agent_id, channel_name))

    def get_channel_named(self, channel_name: str, agent_id: str) -> Optional[T]:
        """Fetch a channel by its name, for a given agent.

        Parameters
        ----------
        channel_name : str
            The unique channel name for the channel.
        agent_id : str
            The unique identifier for the agent that owns the channel.

        Returns
        -------
        Channel, Processor or Task
            The channel object, which can be a Channel, Processor or Task depending on the type of the channel.
        """
        data = self._get_channel_named_raw(channel_name, agent_id)
        return data and self._parse_channel(data)

    def get_channel_messages(
        self, channel_id: str, num_messages: Optional[int] = None
    ) -> list[Message]:
        """Fetch messages from a channel.

        Parameters
        ----------
        channel_id : str
            The unique identifier for the channel.
        num_messages : int, optional
            The number of messages to fetch. If not provided, all messages will be fetched.

        Returns
        -------
        list[Message]
            A list of Message objects from the channel.
        """
        if num_messages:
            data = self.request(
                Route(
                    "GET",
                    "/ch/v1/channel/{}/messages/{}/",
                    channel_id,
                    str(num_messages),
                )
            )
        else:
            data = self.request(Route("GET", "/ch/v1/channel/{}/messages/", channel_id))

        if not data:
            return []

        return [
            Message(client=self, data=m, channel_id=channel_id)
            for m in data["messages"]
        ]

    def _get_message_raw(self, channel_id: str, message_id: str) -> dict[str, Any]:
        return self.request(
            Route("GET", "/ch/v1/channel/{}/message/{}", channel_id, message_id)
        )

    def get_channel_messages_in_window(
        self, channel_id: str, start: datetime, end: datetime
    ) -> list[Message]:
        """Fetch messages from a channel within a specific time window.

        Parameters
        ----------
        channel_id : str
            The unique identifier for the channel.
        start : datetime
            The start and end datetime objects defining the time window.
        end: datetime
            The start and end datetime objects defining the time window.
        """
        start = str(int(start.timestamp()))
        end = str(int(end.timestamp()))
        data = self.request(
            Route(
                "GET", "/ch/v1/channel/{}/messages/time/{}/{}/", channel_id, start, end
            )
        )
        if not data:
            return []

        return [
            Message(client=self, data=m, channel_id=channel_id)
            for m in data["messages"]
        ]

    def get_message(self, channel_id: str, message_id: str) -> Optional[Message]:
        """Fetch a message by its ID from a channel.

        Parameters
        ----------
        channel_id : str
            The unique identifier for the channel.
        message_id : str
            The unique identifier for the message.

        Returns
        -------
        Message
            The message object.

        Raises
        ------
        NotFound
            If the message does not exist in the channel.
        """
        data = self._get_message_raw(channel_id, message_id)
        return data and Message(client=self, data=data, channel_id=channel_id)

    def _delete_message_raw(self, channel_id: str, message_id: str) -> bool:
        return self.request(
            Route("DELETE", "/ch/v1/channel/{}/message/{}", channel_id, message_id)
        )

    def create_channel(self, channel_name: str, agent_id: str) -> T:
        """Create a channel with the given name for the specified agent.

        If the channel already exists, it will return the existing channel.

        Parameters
        ----------
        channel_name : str
            The unique name for the channel. Should start with a '#' for processors or '!' for tasks.
        agent_id : str
            The owner agent's unique identifier.

        Returns
        -------
        Channel, Processor or Task
            The channel object, which can be a Channel, Processor or Task depending on the type of the channel.

        """
        try:
            return self.get_channel_named(channel_name, agent_id)
        except NotFound:
            pass
        # all we need to do is publish to a channel with an empty payload
        self.request(Route("POST", "/ch/v1/agent/{}/{}/", agent_id, channel_name))
        # this is a bit of a wasted API call, but since this is the same method to post an aggregate to a
        # channel it can either return a new channel ID (if created), or the message ID of the posted message.
        return self.get_channel_named(channel_name, agent_id)

    def create_processor(self, processor_name: str, agent_id: str) -> Processor:
        """Create a processor channel with the given name for the specified agent.

        If the processor already exists, it will return the existing processor.

        Parameters
        ----------
        processor_name : str
            The unique name for the processor channel.
        agent_id : str
            The owner agent's unique identifier.

        Returns
        -------
        Processor
            The processor channel object.
        """
        return self.create_channel("#" + processor_name.lstrip("#"), agent_id)

    def create_task(self, task_name: str, agent_id: str, processor_id: str) -> Task:
        """Create a task channel with the given name for the specified agent and an associated processor.

        Parameters
        ----------
        task_name : str
            The unique name for the task channel.
        agent_id : str
            The owner agent's unique identifier.
        processor_id : str
            The processor to associate with this task

        Returns
        -------
        Task
            The task channel object.
        """

        task = "!" + task_name.lstrip("!")
        payload = {
            "msg": {},  # this is a required field apparently
            "processor_id": processor_id,
        }
        self.request(Route("POST", "/ch/v1/agent/{}/{}/", agent_id, task), json=payload)
        return self.get_channel_named(task, agent_id)

    def _maybe_subscribe_to_channel(
        self, channel_id: str, task_id: str, subscribe: bool
    ):
        data = {"channel_id": channel_id, "subscribe": subscribe}
        return self.request(
            Route("POST", "/ch/v1/channel/{}/subscribe/", task_id), json=data
        )

    def subscribe_to_channel(self, channel_id: str, task_id: str) -> bool:
        """Subscribe a task to a channel.

        Parameters
        ----------
        channel_id : str
            The unique identifier for the channel.
        task_id : str
            The unique identifier for the task.

        Returns
        -------
        bool
            True if the subscription was successful, False otherwise.
        """
        return self._maybe_subscribe_to_channel(channel_id, task_id, True)

    def unsubscribe_from_channel(self, channel_id: str, task_id: str) -> bool:
        """Unsubscribe a task from a channel.

        Parameters
        ----------
        channel_id : str
            The unique identifier for the channel.
        task_id : str
            The unique identifier for the task.

        Returns
        -------
        bool
            True if the unsubscription was successful, False otherwise.
        """
        return self._maybe_subscribe_to_channel(channel_id, task_id, False)

    def publish_to_channel(
        self,
        channel_id: str,
        data: Any,
        save_log: bool = True,
        log_aggregate: bool = False,
        override_aggregate: bool = False,
        timestamp: Optional[datetime] = None,
    ):
        """Publish data to a channel.

        Parameters
        ----------
        channel_id : str
            The unique identifier for the channel.
        data : Any
            The data to publish to the channel. This can be a string or a dictionary.
        save_log : bool, optional
            Whether to save the log of the message. Defaults to True.
        log_aggregate : bool, optional
            Whether to aggregate the log of the message. Defaults to False.
        override_aggregate : bool, optional
            Whether to override any existing aggregate with a completely fresh new message
        timestamp : datetime, optional
            The timestamp to set for the message. If not provided, the current time will be used.
        """
        # basically we're assuming there's only 2 types of data - dict or string...
        post_data = {"msg": data}

        post_data["record_log"] = save_log
        if log_aggregate:
            post_data["log_aggregate"] = True
        if override_aggregate:
            post_data["override_aggregate"] = True
        if timestamp:
            post_data["timestamp"] = int(timestamp.timestamp())

        if isinstance(post_data, dict):
            return self.request(
                Route("POST", "/ch/v1/channel/{}/", channel_id), json=post_data
            )
        else:
            return self.request(
                Route("POST", "/ch/v1/channel/{}/", channel_id), data=str(post_data)
            )

    def publish_to_channel_name(
        self,
        agent_id: str,
        channel_name: str,
        data: Any,
        save_log: bool = True,
        log_aggregate: bool = False,
        override_aggregate: bool = False,
        timestamp: Optional[datetime] = None,
    ):
        """Publish data to a channel by its name.

        Parameters
        ----------
        agent_id : str
            The agent ID who owns this channel.
        channel_name : str
            The name for the channel.
        data : Any
            The data to publish to the channel. This can be a string or a dictionary.
        save_log : bool, optional
            Whether to save the log of the message. Defaults to True.
        log_aggregate : bool, optional
            Whether to aggregate the log of the message. Defaults to False.
        override_aggregate : bool, optional
            Whether to override any existing aggregate with a completely fresh new message
        timestamp : datetime, optional
            The timestamp to set for the message. If not provided, the current time will be used.
        """
        post_data = {"msg": data}

        post_data["record_log"] = save_log
        if log_aggregate:
            post_data["log_aggregate"] = True
        if override_aggregate:
            post_data["override_aggregate"] = True
        if timestamp:
            post_data["timestamp"] = int(timestamp.timestamp())

        if isinstance(post_data, dict):
            return self.request(
                Route("POST", "/ch/v1/agent/{}/{}/", agent_id, channel_name),
                json=post_data,
            )
        else:
            return self.request(
                Route("POST", "/ch/v1/agent/{}/{}/", agent_id, channel_name),
                data=str(post_data),
            )

    def create_tunnel_endpoints(self, agent_id: str, endpoint_type: str, amount: int):
        to_return = []
        for i in range(amount):
            res = self.request(
                Route(
                    "POST", "/ch/v1/agent/{}/ngrok_tunnels/{}/", agent_id, endpoint_type
                )
            )
            if res and res.get("url"):
                to_return.append(res["url"])
        return to_return

    def get_tunnel_endpoints(self, agent_id: str, endpoint_type: str):
        return self.request(
            Route("GET", "/ch/v1/agent/{}/ngrok_tunnels/{}/", agent_id, endpoint_type)
        )

    def get_tunnel(self, tunnel_id: str):
        return self.request(Route("GET", "/ch/v1/tunnels/{}/", tunnel_id))

    def get_tunnels(self, agent_id: str, show_choices: bool = False):
        return self.request(
            Route("GET", "/ch/v1/agent/{}/dd_tunnels/", agent_id, choices=show_choices)
        )

    def create_tunnel(self, agent_id: str, **data):
        return self.request(
            Route("POST", "/ch/v1/agent/{}/dd_tunnels/", agent_id), json=data
        )

    def patch_tunnel(self, tunnel_id: str, **data):
        return self.request(Route("PATCH", "/ch/v1/tunnels/{}/", tunnel_id), json=data)

    def activate_tunnel(self, tunnel_id: str):
        return self.request(Route("POST", "/ch/v1/tunnels/{}/activate/", tunnel_id))

    def deactivate_tunnel(self, tunnel_id: str):
        return self.request(Route("POST", "/ch/v1/tunnels/{}/deactivate/", tunnel_id))

    def delete_tunnel(self, tunnel_id: str):
        return self.request(Route("DELETE", "/ch/v1/tunnels/{}/", tunnel_id))

    # applications. only supports operations on apps, not installs / deployments at the moment.
    def get_applications(self):
        """Get the list of applications available to the current agent."""
        return self.request(Route("GET", "/apps/api/v1/applications/"))

    def create_application(
        self, application: Application, is_staging: bool = False
    ) -> str:
        """Create a new application with the given data."""
        payload = application.to_dict(
            include_deployment_data=True, is_staging=is_staging
        )
        data = self.request(Route("POST", "/apps/api/v1/applications/"), json=payload)
        return data["key"]

    def get_application(self, key: str) -> Application:
        """Get a specific application by its key."""
        data = self.request(Route("GET", "/apps/api/v1/applications/{}/", key))
        return Application.from_data(data=data)

    def update_application(
        self, application: Application, is_staging: bool = False
    ) -> None:
        """Update an existing application with the given data."""
        payload = application.to_dict(
            include_deployment_data=True, is_staging=is_staging
        )
        return self.request(
            Route("PATCH", "/apps/api/v1/applications/{}/", payload["key"]),
            json=payload,
        )

    # login methods

    def fetch_token(self):
        """Fetch a temporary token from the cloud API.

        By default, this uses the username and password set in the client, and will fail if this is not set.

        You can override this to implement a custom token refresher, e.g. using the device agent websocket connection.

        This must return a tuple of (token, expires_at, agent_id).
        """
        if not (self.username or self.password):
            raise RuntimeError(
                "Must have username and password set since access token has expired."
            )

        logging.info("Logging in...")
        session = requests.Session()

        login_url = f"{self.base_url}/accounts/login/"

        session.get(login_url)
        login_data = dict(
            login=self.username,
            password=self.password,
            csrfmiddlewaretoken=session.cookies.get("csrftoken"),
            next="/",
        )
        res = session.post(login_url, data=login_data, headers=dict(Referer=login_url))

        # bit of a hack... don't know a better way? Two-Factor is the title on the page...
        if "Two-Factor" in res.text:
            print(
                "Your account has 2FA enabled. It is recommended to instead use `doover configure_token` "
                "and use a long-lived token, otherwise you will have to 2FA authenticate every 20min.\n"
                "Quit and run that command, or supply your 2FA code to authenticate now.\n"
            )

            token = input("Please enter your 2FA token: ")
            twofa_data = dict(
                csrfmiddlewaretoken=session.cookies.get("csrftoken"),
                code=token,
            )
            twofa_url = f"{self.base_url}/accounts/2fa/authenticate/"
            res = session.post(
                twofa_url, data=twofa_data, headers=dict(Referer=twofa_url)
            )
            if res.status_code != 200:
                raise RuntimeError("Failed to authenticate 2FA.")

        res = session.get(f"{self.base_url}/ch/v1/get_temp_token/")

        try:
            data = res.json()
        except requests.exceptions.JSONDecodeError:
            raise RuntimeError("Failed to get temporary token. Login failed.")

        # FIXME: can these expire in UTC?
        difference = timedelta(
            seconds=float(data["valid_until"]) - float(data["current_time"])
        )
        expires_at = datetime.now(timezone.utc) + difference
        return data["token"], expires_at, data["agent_id"]

    def login(self):
        token, expires_at, agent_id = self.fetch_token()
        self._set_login_data(token, expires_at, agent_id)

    def _set_login_data(self, token, expires_at, agent_id):
        self.access_token = AccessToken(token=token, expires_at=expires_at)
        self.agent_id = agent_id
        self.update_headers()

        logging.info(
            f"Successfully logged in and set token to expire "
            f"in {int((expires_at - datetime.now(timezone.utc)).total_seconds() / 60)}min..."
        )
        try:
            self.login_callback()
        except Exception as e:
            print(f"failed to call callback: {e}")
            pass



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import pathlib
import sys

root_project = pathlib.Path(__file__).parents[2].resolve()
sys.path.insert(0, str(pathlib.Path(__file__).resolve().parents[2]))
sys.path.insert(0, root_project.as_posix())
sys.path.append((root_project / "docs").as_posix())

project = "pydoover"
copyright = "2025, Span Engineering Pty Ltd"
author = "Doover Team"
release = "0.4"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
html_theme = "sphinx_rtd_theme"
# html_theme = 'alabaster'
html_static_path = ["_static"]



from pydoover.docker import DockerApplication, run_app

from app_config import SampleConfig


class SampleApplication(DockerApplication):
    config: SampleConfig  # not necessary, but helps your IDE provide autocomplete!

    def setup(self):
        print(f"Hello! The funny config message is: {self.config.funny_message.value}!")

    def main_loop(self):
        if self.config.outputs_enabled.value is True:
            print("We are allowed to set outputs!")

        for i in range(self.config.num_di.value):
            status = self.platform_iface.get_di(i)
            if status:
                print(f"DI {i} is active!")
            else:
                print(f"DI {i} is not active!")


if __name__ == "__main__":
    run_app(SampleApplication(config=SampleConfig()))



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class RequestHeader(_message.Message):
    __slots__ = ("app_id",)
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    app_id: str
    def __init__(self, app_id: _Optional[str] = ...) -> None: ...

class ResponseHeader(_message.Message):
    __slots__ = ("success", "cloud_synced", "cloud_ready", "response_code", "response_message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    CLOUD_SYNCED_FIELD_NUMBER: _ClassVar[int]
    CLOUD_READY_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_CODE_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    cloud_synced: bool
    cloud_ready: bool
    response_code: int
    response_message: str
    def __init__(self, success: bool = ..., cloud_synced: bool = ..., cloud_ready: bool = ..., response_code: _Optional[int] = ..., response_message: _Optional[str] = ...) -> None: ...

class MessageDetails(_message.Message):
    __slots__ = ("message_id", "channel_name", "payload", "timestamp")
    MESSAGE_ID_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_NAME_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    message_id: str
    channel_name: str
    payload: str
    timestamp: str
    def __init__(self, message_id: _Optional[str] = ..., channel_name: _Optional[str] = ..., payload: _Optional[str] = ..., timestamp: _Optional[str] = ...) -> None: ...

class ChannelDetails(_message.Message):
    __slots__ = ("channel_name", "aggregate")
    CHANNEL_NAME_FIELD_NUMBER: _ClassVar[int]
    AGGREGATE_FIELD_NUMBER: _ClassVar[int]
    channel_name: str
    aggregate: str
    def __init__(self, channel_name: _Optional[str] = ..., aggregate: _Optional[str] = ...) -> None: ...

class AgentDetails(_message.Message):
    __slots__ = ("agent_id", "agent_name", "channels")
    AGENT_ID_FIELD_NUMBER: _ClassVar[int]
    AGENT_NAME_FIELD_NUMBER: _ClassVar[int]
    CHANNELS_FIELD_NUMBER: _ClassVar[int]
    agent_id: str
    agent_name: str
    channels: _containers.RepeatedCompositeFieldContainer[ChannelDetails]
    def __init__(self, agent_id: _Optional[str] = ..., agent_name: _Optional[str] = ..., channels: _Optional[_Iterable[_Union[ChannelDetails, _Mapping]]] = ...) -> None: ...

class TestCommsRequest(_message.Message):
    __slots__ = ("header", "message")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    header: RequestHeader
    message: str
    def __init__(self, header: _Optional[_Union[RequestHeader, _Mapping]] = ..., message: _Optional[str] = ...) -> None: ...

class TestCommsResponse(_message.Message):
    __slots__ = ("response_header", "response")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    response: str
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., response: _Optional[str] = ...) -> None: ...

class MessageDetailsRequest(_message.Message):
    __slots__ = ("header", "message_id")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_ID_FIELD_NUMBER: _ClassVar[int]
    header: RequestHeader
    message_id: str
    def __init__(self, header: _Optional[_Union[RequestHeader, _Mapping]] = ..., message_id: _Optional[str] = ...) -> None: ...

class MessageDetailsResponse(_message.Message):
    __slots__ = ("response_header", "message")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    message: MessageDetails
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., message: _Optional[_Union[MessageDetails, _Mapping]] = ...) -> None: ...

class ChannelDetailsRequest(_message.Message):
    __slots__ = ("header", "channel_name")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_NAME_FIELD_NUMBER: _ClassVar[int]
    header: RequestHeader
    channel_name: str
    def __init__(self, header: _Optional[_Union[RequestHeader, _Mapping]] = ..., channel_name: _Optional[str] = ...) -> None: ...

class ChannelDetailsResponse(_message.Message):
    __slots__ = ("response_header", "channel")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    channel: ChannelDetails
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., channel: _Optional[_Union[ChannelDetails, _Mapping]] = ...) -> None: ...

class ChannelSubscriptionRequest(_message.Message):
    __slots__ = ("header", "channel_name")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_NAME_FIELD_NUMBER: _ClassVar[int]
    header: RequestHeader
    channel_name: str
    def __init__(self, header: _Optional[_Union[RequestHeader, _Mapping]] = ..., channel_name: _Optional[str] = ...) -> None: ...

class ChannelSubscriptionResponse(_message.Message):
    __slots__ = ("response_header", "channel", "message")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    channel: ChannelDetails
    message: MessageDetails
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., channel: _Optional[_Union[ChannelDetails, _Mapping]] = ..., message: _Optional[_Union[MessageDetails, _Mapping]] = ...) -> None: ...

class ChannelWriteRequest(_message.Message):
    __slots__ = ("header", "channel_name", "message_payload", "save_log", "max_age")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_NAME_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    SAVE_LOG_FIELD_NUMBER: _ClassVar[int]
    MAX_AGE_FIELD_NUMBER: _ClassVar[int]
    header: RequestHeader
    channel_name: str
    message_payload: str
    save_log: bool
    max_age: int
    def __init__(self, header: _Optional[_Union[RequestHeader, _Mapping]] = ..., channel_name: _Optional[str] = ..., message_payload: _Optional[str] = ..., save_log: bool = ..., max_age: _Optional[int] = ...) -> None: ...

class DebugInfoRequest(_message.Message):
    __slots__ = ("header",)
    HEADER_FIELD_NUMBER: _ClassVar[int]
    header: RequestHeader
    def __init__(self, header: _Optional[_Union[RequestHeader, _Mapping]] = ...) -> None: ...

class ChannelWriteResponse(_message.Message):
    __slots__ = ("response_header", "message_id")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_ID_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    message_id: str
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., message_id: _Optional[str] = ...) -> None: ...

class DebugChannelState(_message.Message):
    __slots__ = ("channel_name", "active", "local_updated_at", "cloud_updated_at")
    CHANNEL_NAME_FIELD_NUMBER: _ClassVar[int]
    ACTIVE_FIELD_NUMBER: _ClassVar[int]
    LOCAL_UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    CLOUD_UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    channel_name: str
    active: bool
    local_updated_at: float
    cloud_updated_at: float
    def __init__(self, channel_name: _Optional[str] = ..., active: bool = ..., local_updated_at: _Optional[float] = ..., cloud_updated_at: _Optional[float] = ...) -> None: ...

class DebugPendingMessage(_message.Message):
    __slots__ = ("channel_name", "app_key", "save_log", "payload", "publish_in")
    CHANNEL_NAME_FIELD_NUMBER: _ClassVar[int]
    APP_KEY_FIELD_NUMBER: _ClassVar[int]
    SAVE_LOG_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    PUBLISH_IN_FIELD_NUMBER: _ClassVar[int]
    channel_name: str
    app_key: str
    save_log: bool
    payload: str
    publish_in: int
    def __init__(self, channel_name: _Optional[str] = ..., app_key: _Optional[str] = ..., save_log: bool = ..., payload: _Optional[str] = ..., publish_in: _Optional[int] = ...) -> None: ...

class DebugInfoResponse(_message.Message):
    __slots__ = ("response_header", "active_callbacks", "wss_callbacks", "wss_aggregates", "wss_subscriptions", "channels", "pending_messages", "wss_connected")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    ACTIVE_CALLBACKS_FIELD_NUMBER: _ClassVar[int]
    WSS_CALLBACKS_FIELD_NUMBER: _ClassVar[int]
    WSS_AGGREGATES_FIELD_NUMBER: _ClassVar[int]
    WSS_SUBSCRIPTIONS_FIELD_NUMBER: _ClassVar[int]
    CHANNELS_FIELD_NUMBER: _ClassVar[int]
    PENDING_MESSAGES_FIELD_NUMBER: _ClassVar[int]
    WSS_CONNECTED_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    active_callbacks: _containers.RepeatedScalarFieldContainer[str]
    wss_callbacks: _containers.RepeatedScalarFieldContainer[str]
    wss_aggregates: _containers.RepeatedScalarFieldContainer[str]
    wss_subscriptions: _containers.RepeatedScalarFieldContainer[str]
    channels: _containers.RepeatedCompositeFieldContainer[DebugChannelState]
    pending_messages: _containers.RepeatedCompositeFieldContainer[DebugPendingMessage]
    wss_connected: bool
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., active_callbacks: _Optional[_Iterable[str]] = ..., wss_callbacks: _Optional[_Iterable[str]] = ..., wss_aggregates: _Optional[_Iterable[str]] = ..., wss_subscriptions: _Optional[_Iterable[str]] = ..., channels: _Optional[_Iterable[_Union[DebugChannelState, _Mapping]]] = ..., pending_messages: _Optional[_Iterable[_Union[DebugPendingMessage, _Mapping]]] = ..., wss_connected: bool = ...) -> None: ...

class TempAPITokenRequest(_message.Message):
    __slots__ = ("header",)
    HEADER_FIELD_NUMBER: _ClassVar[int]
    header: RequestHeader
    def __init__(self, header: _Optional[_Union[RequestHeader, _Mapping]] = ...) -> None: ...

class TempAPITokenResponse(_message.Message):
    __slots__ = ("response_header", "token", "valid_until", "endpoint")
    RESPONSE_HEADER_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    VALID_UNTIL_FIELD_NUMBER: _ClassVar[int]
    ENDPOINT_FIELD_NUMBER: _ClassVar[int]
    response_header: ResponseHeader
    token: str
    valid_until: str
    endpoint: str
    def __init__(self, response_header: _Optional[_Union[ResponseHeader, _Mapping]] = ..., token: _Optional[str] = ..., valid_until: _Optional[str] = ..., endpoint: _Optional[str] = ...) -> None: ...
